"""This is the demo script to show how to work with Scripts and execute
Python code server-side in Strategy One.

This script showcases basic Code and Script management, execution, and
variable handling. It will not work without replacing parameters with
real values. Its goal is to present what can be done with this module
and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.python_execution.script import (
    Code,
    Script,
    ScriptEnvironmentError,
    ScriptError,
    ScriptExecutionError,
    ScriptSetupError,
    VariableType,
    list_scripts,
    VariableStandardScript,
    VariableDatasourceScript,
    VariableTransactionScript,
    VariableAnswer,
    ExecutionStatus,
    ScriptResultType,
    ScriptUsageType,
)
from pathlib import Path

# Define variables which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here
PROJECT_ID = $project_id  # Insert project ID here
SCRIPT_NAME = $script_name  # Insert name of the Script here
SCRIPT_ID = $script_id  # Insert ID of Script here
RUNTIME_ID = $runtime_id  # Insert ID of runtime here
FOLDER_ID = $folder_id  # Insert folder ID here

# Create connection based on workstation data
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# --- Code Execution ---
# Create a Code instance from a string
code = Code(conn, code="print('Hello from Strategy One')")  # will not validate the code
code = Code(conn, code="print('Hello from Strategy One')", validate_code=True)

# Load code from a file (with or without validation)
code = Code.get_from_file(conn, Path('./my_script.py'))  # will not validate the code
code = Code.get_from_file(conn, Path('./my_script.py'), validate_code=True)
# or path as string
code = Code.get_from_file(conn, './my_script.py', validate_code=True)

# Execute code synchronously and wait for completion
status = code.execute()
print(f"Execution status: {status}")

# Get execution console output
output = code.stdout
print(f"Output: {output}")

# Execute code asynchronously
code.execute(block_until_done=False)
print("Execution initiated.")

# Wait for asynchronous execution to finish
final_status = code.wait_for_execution_finish()
print(f"Final status: {final_status}")

# Get execution details
details = code.get_execution_details()
print(f"Execution details: {details}")

# Stop running code execution
code.execute(block_until_done=False)
code.stop_execution()
status = code.status
print(f"Code execution stopped: {status}")

# --- Code with Variables ---
# Create code that uses variables (denoted with $variable_name)
code_with_var = Code(
    conn,
    code="result = $input_value * 2\nprint(result)",
    validate_code=True,
)

# Execute code with variable values
status = code_with_var.execute(
    variables=[
        VariableStandardScript(
            'input_value', value=21, type=VariableType.NUMERICAL, prompt=False
        )
    ]
)

# Execute with interactive prompting for variables
status = code_with_var.execute(
    variables=[
        VariableStandardScript(
            'input_value', value=21, type=VariableType.NUMERICAL, prompt=True
        )
    ],
    answers={'input_value': '42'},
)

# Pipe logs to see execution output
status = code_with_var.execute(pipe_logs=True)

# Raise exception if code execution fails
try:
    code_which_errors = Code(conn, code="raise Exception('Error!')")
    status = code_which_errors.execute(raise_on_execution_failure=True)
except ScriptError as err:
    print(f"Code execution failed: {err}")

# Read properties of last execution status
print(code.stdout)  # When Code is successful, STDOUT of the Code
print(code.stderr)  # When Code is successful, STDERR of the Code
print(code.executor_message)  # When Code errored out during runtime, error details of why

# --- Script Management ---
# List all scripts in a project
all_scripts = list_scripts(
    connection=conn,
    project=PROJECT_ID,
    to_dictionary=False,
)

# List scripts with filtering
filtered_scripts = list_scripts(
    connection=conn,
    project=PROJECT_ID,
    name=SCRIPT_NAME,
    to_dictionary=False,
)

# Get script as dictionaries instead of objects
scripts_as_dicts = list_scripts(
    connection=conn,
    project=PROJECT_ID,
    to_dictionary=True,
)

# Get single script by ID
script = Script(connection=conn, id=SCRIPT_ID)

# Get single script by name
script = Script(connection=conn, name=SCRIPT_NAME)

# List script properties
properties = script.list_properties()
print(properties)

# --- Script Creation ---
# Create a simple script
new_script = Script.create(
    connection=conn,
    name='MyScript',
    runtime_id=RUNTIME_ID,
    code='print("Hello")',  # will not be validated
    destination_folder=FOLDER_ID,
)

# Create script with validation
new_script = Script.create(
    connection=conn,
    name='ValidatedScript',
    runtime_id=RUNTIME_ID,
    code='x = 42\nprint(x)',
    destination_folder=FOLDER_ID,
    validate_code=True,  # will validate code before creation
)

# Create script with variables
new_script_with_vars = Script.create(
    connection=conn,
    name='ScriptWithVariables',
    runtime_id=RUNTIME_ID,
    code='print($name)\nprint($age)',
    destination_folder=FOLDER_ID,
    variables=[
        # `VariableStandardScript` is by default prompted unless set otherwise
        VariableStandardScript('name', value='John'),
        VariableStandardScript('age', value=30, type=VariableType.NUMERICAL),
        VariableStandardScript('email'),
        VariableStandardScript('title', value='Junior', prompt=False),
        # [Note]: There are other variable-defining classes available, specific
        # to other script usage types. Please see their respective docstrings
        # for more details
    ],
    validate_code=True,
)

# Create script with result type
new_script_with_result = Script.create(
    connection=conn,
    name='ScriptWithResult',
    runtime_id=RUNTIME_ID,
    # see `script-templates` to check how to return from a returning script
    code='def get_results():\n    return 42',
    destination_folder=FOLDER_ID,
    script_result_type=ScriptResultType.NUMERICAL,  # requires `get_results` to be defined
    validate_code=True,
)

# Create Datasource script instead of a standard one
new_datasource_script = Script.create(
    connection=conn,
    name='MyDatasourceScript',
    runtime_id=RUNTIME_ID,
    code='data = [1, 2, 3]\n...',  # this is only mocked code, replace with a valid one
    destination_folder=FOLDER_ID,
    script_usage_type=ScriptUsageType.DATASOURCE,
    variables=[
        VariableDatasourceScript('filter_param', value='default'),
    ],
    validate_code=True,
)

# --- Script Execution ---
# Execute script synchronously
status = script.execute()
print(f"Script execution status: {status}")

# Get script execution console output
output = script.execution_stdout
print(f"Script output: {output}")

# Execute script asynchronously
script.execute(block_until_done=False)
# [Note]: no status check here as with this flag off, it is an asynchronous operation

# Wait for script execution
final_status = script.wait_for_execution_finish()
print(f"Script execution status: {final_status}")

# Execute with variable answers
status = new_script_with_vars.execute(
    variables_answers={
        'name': 'Charlie',
        # this flag allows to keep a default Variable value as an answer to its
        # prompt for this execution
        'age': VariableAnswer.KEEP_GLOBAL_DEFAULT,
        # OR
        # If previously set, this flag allows to keep a default Variable
        # personal answer as an answer to its prompt for this execution
        'email': VariableAnswer.KEEP_PERSONAL_DEFAULT,
    },
)

# Execute with variable answers using VariableAnswer class, for code readability
status = new_script_with_vars.execute(
    variables_answers=[
        VariableAnswer.For('name').should_be('Bob'),
        # this flag allows to keep a default Variable value as an answer to its
        # prompt for this execution
        VariableAnswer.For('age').should_be_global_default,
        # OR
        # If previously set, this flag allows to keep a default Variable
        # personal answer as an answer to its prompt for this execution
        VariableAnswer.For('email').should_be_personal_default,
    ],
)

# execute script synchronously allowing to answer variables on the fly
# [Disclaimer]: Does not work in Workstation, only standalone
status = new_script_with_vars.execute(
    allow_interactive_answering=True,
)
# [Note]: you will be prompted to provide answer to each prompted unanswered
# variable, one by one

print(f"Script execution status: {status}")

# Get script execution status
current_status = script.get_execution_status()

# Get detailed execution information
execution_details = script.get_execution_details()
print(execution_details)

# Get last execution run details
last_run_details = script.get_last_run_details()

# Stop running script execution
script.execute(block_until_done=False)
script.stop_execution()

status = script.execution_status
print(f"Script execution stopped: {status}")

# --- Script Variables Management ---
# Get script variables
variables = script.get_variables()

# Add variables to script
script.add_variables(
    VariableStandardScript('new_var', value='default'),  # using dedicated Variable class
    {'name': 'another_var', 'type': VariableType.TEXT},  # using dictionary of data
)

# Alter existing variables
script.alter_variables(
    changes={
        'new_var': VariableStandardScript('new_var', value='updated'),  # using Variable class
        'another_var': {'value': 'new_default'},  # using dictionary of data
    }
)

# Delete variables from script
script.delete_variables(
    'new_var',  # by name
    VariableStandardScript("another_var")  # by Variable object
)

# Save personal variable answers (for prompted variables)
script.save_personal_variable_answers(
    answers={'prompt_var': 'my_answer'}
)

# --- Script Modification ---
# Alter script properties
script.alter(
    name='UpdatedScriptName',
    description='Updated description',
    code='print("Updated code")',
    validate_code=True,
)

# Change script runtime
script.alter(runtime_id=RUNTIME_ID)

# Change script result type
script.alter(script_result_type=ScriptResultType.TEXT)

# Update all variables in script at once
script.alter(
    variables=[
        VariableStandardScript('var1', value='val1'),
        VariableStandardScript('var2', value='val2'),
    ]
)

# --- Script Code Access ---
# Get script code as text
code_text = script.code
print(code_text)

# Read properties of last execution status
print(script.execution_stdout)  # When Script is successful, STDOUT of the Script
print(script.execution_stderr)  # When Script is successful, STDERR of the Script
print(script.execution_pod_executor_message)  # When Script errored out during runtime, error details of why
print(script.execution_result)  # When Script is successful and with return type, the returned result

# --- Script Deletion ---
# Delete script with confirmation prompt
script.delete()

# Delete script without confirmation prompt
script.delete(force=True)

# --- Datasource and Transaction Scripts ---
# Create a datasource script
new_datasource_script = Script.create(
    connection=conn,
    name='DatasourceScript',
    runtime_id=RUNTIME_ID,
    code='data = [1, 2, 3]\nprint(data)',
    destination_folder=FOLDER_ID,
    script_usage_type=ScriptUsageType.DATASOURCE,
    validate_code=True,
)

# Create a transaction script with transaction column variable
transaction_script = Script.create(
    connection=conn,
    name='TransactionScript',
    runtime_id=RUNTIME_ID,
    code='print($txn_var)',
    destination_folder=FOLDER_ID,
    script_usage_type=ScriptUsageType.TRANSACTION,
    variables=[
        VariableTransactionScript(
            'txn_var',
            transaction_column=True,
            prompt=True,
            multiple=True,
        )
    ],
    validate_code=True,
)

# Execute datasource script
status = new_datasource_script.execute()

# Execute transaction script with answers
status = transaction_script.execute(
    variables_answers={'txn_var': ['value1', 'value2']},
)

# --- Error Handling ---
# Handle execution errors
try:
    code_with_error = Code(conn, code="1 / 0")
    status = code_with_error.execute(raise_on_execution_failure=True)
except ScriptError as err:
    print(f"Execution error: {err}")

# Read execution error when logs are not piped
code_with_error = Code(conn, code="1 / 0")
status = code_with_error.execute()
if status == ExecutionStatus.ERROR:
    error_message = code_with_error.stderr
    print(f"Execution failed with error: {error_message}")

# Check if code is valid
if code.is_valid():
    print("Code is valid")
else:
    print("Code has syntax errors")



from typing import TYPE_CHECKING

from mstrio.api import change_journal as change_journal_api
from mstrio.utils.helper import camel_to_snake

if TYPE_CHECKING:
    from mstrio.connection import Connection


def get_change_journals_loop(
    connection: 'Connection',
    search_id: str,
    offset: int = 0,
    limit: int = -1,
    fields: str | None = None,
) -> list[dict]:
    """Fetch all change journal search results or up to the specified limit.

    Args:
        connection: Connection object to Strategy environment.
        search_id: ID of the change journal search to retrieve results for.
        offset: Starting offset for pagination.
        limit: Maximum number of results to return. If -1, fetch all results.
        fields: Comma-separated list of fields to include in response.

    Returns:
        List of change journal entry dictionaries.
    """
    limit_per_request = 100 if limit == -1 else min(limit, 100)

    response = change_journal_api.get_change_journal_search_results(
        connection, search_id, offset, limit_per_request, fields
    )
    new_entries = camel_to_snake(response.json().get('changeJournalEntries', []))
    all_entries = new_entries

    if limit != -1 and len(all_entries) >= limit:
        return all_entries[:limit]

    offset = 0
    while len(new_entries) == limit_per_request:
        offset += limit_per_request

        if limit != -1:
            remaining_needed = limit - len(all_entries)
            if remaining_needed <= 0:
                break
            current_limit = min(remaining_needed, 100)
        else:
            current_limit = 100

        response = change_journal_api.get_change_journal_search_results(
            connection, search_id, offset, current_limit, fields
        )
        results = response.json()
        new_entries = camel_to_snake(results.get('changeJournalEntries', []))
        all_entries += new_entries

    if limit != -1:
        return all_entries[:limit]

    return all_entries



import ast
import contextlib
import json
import logging
import re
import time
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum, auto
from inspect import isfunction
from operator import xor
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypeVar, overload

from mstrio import config
from mstrio.helpers import is_valid_datetime, try_str_to_num
from mstrio.object_management.folder import Folder
from mstrio.object_management.search_enums import SearchDomain, SearchPattern
from mstrio.object_management.search_operations import full_search
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.encoder import Encoder
from mstrio.utils.entity import (
    CertifyMixin,
    ChangeJournalMixin,
    CopyMixin,
    DeleteMixin,
    Entity,
    MoveMixin,
)
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import Dictable, delete_none_values
from mstrio.utils.resolvers import (
    get_folder_id_from_params_set,
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import scripts as scripts_processor

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server.project import Project


logger = logging.getLogger(__name__)


# Sync below with any `isinstance` checks done in this module
_VariableValueRaw: TypeAlias = str | int | float | datetime | date
VariableValue: TypeAlias = _VariableValueRaw | list[_VariableValueRaw]
CodeResult: TypeAlias = str | int | float | datetime | None
VariablesAnswers: TypeAlias = (
    list['VariableAnswer | FlagKeepDefaultAnswer | AnswerForCallback']
    # FYI: dict shape {"id-or-name-of-variable": "some-value", ...}
    | dict[str, 'VariableValue | FlagKeepDefaultAnswer']
)
AnswerForCallback: TypeAlias = Callable[[list['Variable']], None]
FlagKeepDefaultAnswer: TypeAlias = (
    type['VariableAnswer.KEEP_PERSONAL_DEFAULT']
    | type['VariableAnswer.KEEP_GLOBAL_DEFAULT']
)


# --- Exceptions ---
class ScriptError(Exception):
    """Base Exception class for other exceptions in Scripts module.

    Created for ease of catching all errors from this module, regardless of
    subtype. All other Exceptions in this module will inherit from it.
    """


class ScriptExecutionError(ScriptError):
    """Exception raised when a script execution fails."""


class ScriptEnvironmentError(ScriptExecutionError):
    """Exception raised when there is an environment-related error during
    script execution.
    """


class ScriptSetupError(ScriptError):
    """Exception raised when a script is not properly set up for execution
    or saving.
    """


# --- END: Exceptions ---


# --- Enums ---
class ExecutionStatus(Enum):
    INITIATED = 0
    RUNNING = 1
    PARTIALLY_READY = 2
    READY = 3  # This and all below are considered "done"
    ERROR = 4
    EXPIRED = 5
    CANCELLED = 6

    @classmethod
    def is_done(cls, value: 'ExecutionStatus | int') -> bool:
        return get_enum_val(value, ExecutionStatus) >= ExecutionStatus.READY.value

    @classmethod
    def is_error(cls, value: 'ExecutionStatus | int') -> bool:
        return get_enum_val(value, ExecutionStatus) >= ExecutionStatus.ERROR.value


class ScriptType(Enum):
    RESERVED = "reserved"
    PYTHON = "python"
    JUPYTER_NOTEBOOK = 'jupyterNotebook'


class ScriptUsageType(Enum):
    STANDARD = 'commandManager'
    DATASOURCE = 'dataSource'
    TRANSACTION = 'transaction'
    # script-as-metric related types below
    SIMPLE_FUNCTION = 'pythonSimpleFunction'
    AGGREGATION_FUNCTION = 'pythonAggFunction'
    AGGREGATION_FUNCTION_NO_SORT = 'pythonNoSortAggFunction'
    RELATIVE_FUNCTION = 'pythonRelativeFunction'
    RELATIVE_FUNCTION_NO_SORT = 'pythonNoSortRelativeFunction'


class ScriptResultType(AutoName):
    """Type of Script's output"""

    NUMERICAL = auto()
    TEXT = auto()
    DATE = auto()
    RESERVED = auto()
    UNKNOWN = auto()


class VariableType(Enum):
    RESERVED = 0x00
    TEXT = 0x01
    NUMERICAL = 0x02
    DATE = 0x03
    SECRET = 0x04
    DATETIME = 0x05
    SYSTEM_PROMPT = 0x100A  # 4106
    TXN_ROW_PROVENANCE = 0x8001  # 32769


# --- END: Enums ---


# --- Configs & Behavior Customization ---
@dataclass
class _VariableAnswerConfig:
    INTERACTIVE_KEEP_PERSONAL_DEFAULT: str = '...'
    INTERACTIVE_KEEP_GLOBAL_DEFAULT: str = '^'


VariableAnswerConfig = _VariableAnswerConfig()
# --- END: Configs & Behavior Customization ---


def list_scripts(
    connection: 'Connection',
    name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    folder: 'Folder | tuple[str] | list[str] | str | None' = None,
    folder_id: str | None = None,
    folder_name: str | None = None,
    folder_path: tuple[str] | list[str] | str | None = None,
    **filters,
) -> 'list[dict] | list[Script]':
    """Get a list of scripts.

    Args:
        connection (Connection): Strategy connection object returned by
            `connection.Connection()`.
        name (str, optional): Name or part of the name of the Script.
        search_pattern (SearchPattern | int, optional): Search pattern for
            finding the Script by name. Defaults to
            `SearchPattern.CONTAINS`.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns Script objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        folder (Folder | tuple | list | str, optional): Folder object or ID or
            name or path specifying the folder. May be used instead of
            `folder_id`, `folder_name` or `folder_path`.
        folder_id (str, optional): ID of a folder.
        folder_name (str, optional): Name of a folder.
        folder_path (str, optional): Path of the folder.
            The path has to be provided in the following format:
                if it's inside of a project, start with a Project Name:
                    /MicroStrategy Tutorial/Public Objects/Metrics
                if it's a root folder, start with `CASTOR_SERVER_CONFIGURATION`:
                    /CASTOR_SERVER_CONFIGURATION/Users
        **filters: Available filter parameters: ['id', 'description, 'owner',
            'date_created', 'date_modified', 'version', 'acg']

    Returns:
        List of scripts as dicts or as Script objects.
    """

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )
    validate_owner_key_in_filters(filters)

    with connection.temporary_project_change(project=proj_id):
        return full_search(
            connection=connection,
            project=proj_id,
            object_types=ObjectTypes.SCRIPT,
            name=name,
            pattern=search_pattern,
            domain=SearchDomain.PROJECT,
            to_dictionary=to_dictionary,
            limit=limit,
            root=folder,
            root_id=folder_id,
            root_name=folder_name,
            root_path=folder_path,
            **filters,
        )


class Code:
    """Class representing Python code to be run in server-side execution
    engine of Strategy.

    Note:
        To retrieve code-as-text from an instance of Code class, just use
        `.as_text` property or stringify the `Code` instance with `str`:
        ```
        >>> code = Code(conn, "print('Hello World')")
        >>> txt = str(code)
        >>> txt2 = code.as_text  # both equal "print('Hello World')"
        ```
    """

    _evaluation_id: str | None = None
    _run_details: dict = {}
    _source_script: 'Script | None' = None
    __validation_error: Exception | None = None

    def __init__(
        self,
        connection: 'Connection',
        code: 'Code | str',
        validate_code: bool = False,
    ) -> None:
        """Initialize Python Code as string or from another Code class instance.

        Args:
            code (Code | str): Other Code class instance or code as a string
                (either already base64-encoded or original)
            validate_code (bool, optional): Whether to validate the code
                syntax upon creation. Defaults to False. (This is a pre-run
                validation only and does not guarantee no Runtime Errors.)
        """

        self._connection = connection

        if isinstance(code, Code):
            code = code._code.decoded_text

        if not code or not isinstance(code, str):
            raise ScriptSetupError(
                "Cannot obtain string version of provided `code` parameter or "
                "the `code` is empty."
            )

        self._code = (
            Encoder(encoded_text=code)
            if Encoder.is_encoded(code)
            else Encoder(decoded_text=code)
        )

        if validate_code:
            self._validate()

    def _validate(self) -> None:
        if not self.is_valid():
            raise ScriptSetupError(
                f"Invalid Code: {str(self)}"
            ) from self.__validation_error

    def _convert_variables_and_answers_to_lists(
        self,
        variables: 'list[Variable | dict] | None',
        answers: VariablesAnswers | None,
        variables_factory: type['Variable'] | None,
    ) -> tuple[list['Variable'], list['VariableAnswer']]:
        """Parses Variables versus their answers to return all variables and
        their answers in common format -> as lists.

        - Variables as a list of `Variable` instances
        - Answers as a list of `VariableAnswer` instances

        Returns:
            `tuple(variables, answers)`
        """

        # FYI: Either explicitly provided factory...
        # OR factory calculated from variables if it's unique for all...
        # OR Standard Script one
        Factory = variables_factory or (  # NOSONAR # (name format)
            options.pop()
            if variables
            and len(
                options := {type(var) for var in variables if isinstance(var, Variable)}
            )
            == 1
            else VariableStandardScript
        )

        variables = variables or []
        answers = answers or []

        if not all(isinstance(var, (dict, Factory)) for var in variables):
            raise ScriptSetupError(
                "Variables have incorrect shape. All of them need to be either "
                "a dictionary of data or an instance of provided factory class "
                f"'{Factory.__name__}'."
            )

        # TODO: create helper method for this "Object | dict -> Object" logic
        variables: list[Variable] = [
            Factory.from_dict(v) if isinstance(v, dict) else v for v in variables
        ]

        if isinstance(answers, dict):
            # convert complex dict to `VariableAnswer` class instances
            try:
                answers = [
                    Factory.get_variable_by_identifier(variables, key).answer(answer)
                    for key, answer in answers.items()
                ]
            except (KeyError, AttributeError):
                raise ScriptSetupError(
                    "Some provided answers refer to non-existent Variables."
                )

        variables = variables.copy()
        answers = answers.copy()

        # calculate answers if done via `VariableAnswer.For(...)...` callback
        for i, answer in enumerate(answers):
            if isfunction(answer) and getattr(answer, '_is_internal_mstrio', None):
                answers[i] = answer(variables)

        return (variables, answers)

    @overload
    def execute(
        self,
        runtime_id: str | None = None,
        variables: 'list[Variable | dict] | None' = None,
        answers: VariablesAnswers | None = None,
        block_until_done: Literal[True] = True,
        pipe_logs: bool = False,
        raise_on_execution_failure: bool = False,
        allow_interactive_answering: bool = False,
        variables_factory: type['Variable'] = None,
    ) -> ExecutionStatus: ...

    @overload
    def execute(
        self,
        runtime_id: str | None = None,
        variables: 'list[Variable | dict] | None' = None,
        answers: VariablesAnswers | None = None,
        block_until_done: Literal[False] = False,
        allow_interactive_answering: bool = False,
        variables_factory: type['Variable'] = None,
    ) -> None: ...

    def execute(
        self,
        # TODO: extend with proper resolver during Module for Runtimes FE dev
        runtime_id: str | None = None,
        variables: 'list[Variable | dict] | None' = None,
        answers: VariablesAnswers | None = None,
        block_until_done: bool = True,
        pipe_logs: bool = False,
        raise_on_execution_failure: bool = False,
        allow_interactive_answering: bool = False,
        # FYI: `None` represents `VariableStandardScript`
        variables_factory: type['Variable'] = None,
    ) -> 'ExecutionStatus | None':
        """Start Code execution.

        Args:
            runtime_id (str, optional): ID of the Script runtime. If nothing is
                provided, defaults to Default Runtime.
            variables (list[_Variable | dict], optional): List of Variables to
                be used during Code execution.
            answers (VariablesAnswers, optional): Answers to prompted Variables.
                Can be either a list of `VariableAnswer` class instances,
                `FlagKeepDefaultAnswer` flags or a dict with shape:
                `{"variable-id-or-name": "answer-value", ...}`.
            block_until_done (bool, optional): Whether to block the execution
                until it is done. Defaults to True.
            pipe_logs (bool, optional): Whether to pipe the execution logs
                to the logger while waiting for the execution to finish. Valid
                only if `block_until_done` is set. Defaults to False.
            raise_on_execution_failure (bool, optional): Whether to raise an
                exception if the execution fails. Valid only if
                `block_until_done` is set. Defaults to False.
            allow_interactive_answering (bool, optional): Whether to allow
                interactive answering of Variables if not all are answered.
                Defaults to False.
            variables_factory (type[Variable], optional): Factory class to use
                for creating Variable instances. Defaults to
                `VariableStandardScript`.

        Returns:
            ExecutionStatus: if `block_until_done` is set to True.
            Nothing otherwise.
        """

        if self._evaluation_id:
            raise ScriptExecutionError(
                "The code is already running. Make sure you wait for its "
                "finalization before trying to run it again or create a new "
                "instance of Script or Code class to run it independently."
            )

        variables, answers = self._convert_variables_and_answers_to_lists(
            variables=variables,
            answers=answers,
            variables_factory=variables_factory,
        )

        if allow_interactive_answering:
            if self._connection.is_run_in_workstation():
                raise ScriptEnvironmentError(
                    "Cannot allow interactive Variables answering in a script "
                    "run in Workstation."
                )

            left_to_consider = tuple(v for v in variables if not v.is_answered())

            if left_to_consider:
                logger.info(
                    f"(Write `{VariableAnswerConfig.INTERACTIVE_KEEP_PERSONAL_DEFAULT}`"
                    " to keep personal default answer or "
                    f"`{VariableAnswerConfig.INTERACTIVE_KEEP_GLOBAL_DEFAULT}` to "
                    "keep global default answer)"
                )

            for var_to_answer in left_to_consider:
                user_input = input(f"Answer to Variable '{var_to_answer.name}': ")

                match user_input:
                    case VariableAnswerConfig.INTERACTIVE_KEEP_GLOBAL_DEFAULT:
                        user_input = VariableAnswer.KEEP_GLOBAL_DEFAULT
                    case VariableAnswerConfig.INTERACTIVE_KEEP_PERSONAL_DEFAULT:
                        user_input = VariableAnswer.KEEP_PERSONAL_DEFAULT
                    case _:
                        if var_to_answer.type == VariableType.NUMERICAL:
                            with contextlib.suppress(ValueError):
                                # keep as string if not a valid float,
                                # `answer` will handle errors
                                user_input = float(user_input)

                var_to_answer.answer(user_input)

        # FYI: raises at Variable that is not properly answered
        [v.validate_whether_answered() for v in variables]

        if not block_until_done and (raise_on_execution_failure or pipe_logs):
            logger.warning(
                "`raise_on_execution_failure` or `pipe_logs` flags will be ignored as "
                "`block_until_done` is not set, therefore the Code will "
                "run in the background, independently."
            )

        if self._source_script:
            self._evaluation_id = scripts_processor.start_run(
                connection=self._source_script.connection,
                script_id=self._source_script.id,
                variables_data=[v.get_answer_as_dict() for v in variables if v.prompt],
            )
        else:
            self._evaluation_id = scripts_processor.start_run(
                connection=self._connection,
                encoded_code=self._code.encoded_text,
                runtime_id=runtime_id,
                variables_data=[
                    v.get_answer_as_dict_for_non_script_run() for v in variables
                ],
            )

        if not self._evaluation_id:
            raise ScriptEnvironmentError("Evaluation ID was not received.")

        if block_until_done:
            ret = self.wait_for_execution_finish(pipe_logs=pipe_logs)

            if raise_on_execution_failure and ExecutionStatus.is_error(ret):
                msg_prefix = (
                    (
                        f"Execution of Script {self._source_script.name} with "
                        f"ID '{self._source_script.id}'"
                    )
                    if self._source_script
                    else "Code execution"
                )

                raise ScriptExecutionError(
                    f"{msg_prefix} failed. Details: {str(self._run_details)}"
                )

            return ret

    def stop_execution(self) -> None:
        """Stops the code execution."""

        if not self._evaluation_id:
            raise ScriptSetupError(
                "Code execution cannot be stopped as it is not running."
            )

        scripts_processor.stop_run(self._connection, self._evaluation_id)
        self.get_execution_details()
        self._evaluation_id = None

    def get_current_execution_status(self) -> ExecutionStatus | None:
        """Retrieve current status of the code execution.

        Returns:
            ExecutionStatus: Current status of the code execution, or None if
                the code is not currently running.
        """

        if not self._evaluation_id:
            return None

        raw_status = self.get_execution_details().get('status')
        return ExecutionStatus(raw_status) if raw_status is not None else None

    def wait_for_execution_finish(
        self, pipe_logs: bool = False, interval: int | None = None
    ) -> ExecutionStatus:
        """Wait until the code execution is finished.

        Args:
            pipe_logs (bool, optional): Whether to pipe the execution logs of
                the Code to the console of this requester script. Defaults to
                False.
            interval (int, optional): Time interval in seconds between polling
                requests for status. If not provided, the value is taken from
                mstrio-py's `config`.

        Returns:
            ExecutionStatus: Final status of the code execution.
        """

        if not self._evaluation_id:
            self.get_execution_details()  # FYI: will throw

        try:
            while not ExecutionStatus.is_done(self.get_current_execution_status()):
                time.sleep(interval or config.delay_between_polling)

        except BaseException as err:
            # At this point the code run on I-Server pod and this requester
            # script should be synchronized. If something happens with this
            # script, we should do something similar to the code run on
            # I-Server as well
            with contextlib.suppress(Exception):
                self.stop_execution()

            raise err

        finally:
            if pipe_logs:
                prefix = "Script" if self._source_script else "Code"

                if self.stdout:
                    logger.info(f"[{prefix} Log]: {self.stdout}")

                if self.stderr:
                    logger.warning(f"[{prefix} Error Log]: {self.stderr}")

                logger.info(f"[{prefix} Completion -> {self.status}]")

        self._evaluation_id = None

        return self.status

    def get_execution_details(self) -> dict:
        """Retrieve details about the current code execution.

        Returns:
            dict: Details about the current code execution.
        """

        if not self._evaluation_id:
            raise ScriptSetupError(
                "Cannot retrieve execution information for non-running code."
            )

        self._run_details = scripts_processor.get_run_result(
            self._connection, self._evaluation_id
        )

        return self._run_details

    def is_valid(self) -> bool:
        """Check whether the code syntax is valid.

        Note:
            This is a pre-run validation only and does not guarantee no
            Runtime Errors.

        Returns:
            bool: True if the code syntax is valid, including Strategy-specific
                extensions, like Variables. False otherwise.
        """

        code = str(self)

        try:
            if "$" in code:
                if re.match(r'\$[a-zA-Z0-9_]+\s*=[^=]', code):
                    # assignment to variable
                    raise SyntaxError(
                        # FYI: content based on what I-Server raises
                        "Strategy Variable cannot be assigned to in Python Code. "
                        "Did you mean to use `==` instead of `=`?"
                    )

                # replace all Variables with a placeholder that is a valid
                # Python constant to check if in that form the code is valid
                code = "__mstrio_vars__ = None\n" + re.sub(
                    r'\$[a-zA-Z0-9_]+', '__mstrio_vars__', code
                )

            compile(
                code,
                "<string>",
                'exec',
                flags=ast.PyCF_TYPE_COMMENTS,
                dont_inherit=True,
                optimize=0,
            )

            return True
        except (SyntaxError, TypeError) as err:
            self.__validation_error = err
            return False

    @property
    def status(self) -> ExecutionStatus | None:
        """Status from last made current execution status request."""

        raw_value = self._run_details.get('status')
        return ExecutionStatus(raw_value) if raw_value is not None else None

    @property
    def stdout(self) -> str:
        """STDOUT from last made current execution status request."""

        return self._run_details.get('results', {}).get('stdout', '').strip()

    @property
    def stderr(self) -> str:
        """STDERR from last made current execution status request."""

        return self._run_details.get('results', {}).get('stderr', '').strip()

    @property
    def executor_message(self) -> str | None:
        """Message from Python Executor (if any) in last made current execution
        status request.

        Usually populated with error details when Code failed during runtime.
        """

        return self._run_details.get('message')

    @property
    def result(self) -> CodeResult:
        """Returned Value from last made current execution status request."""

        ret = json.loads(self._run_details.get('results', {}).get('output', 'null'))

        if self._source_script and ret is not None:
            if self._source_script.script_result_type == ScriptResultType.DATE:
                ret = datetime.fromisoformat(ret)
            elif self._source_script.script_result_type == ScriptResultType.NUMERICAL:
                ret = try_str_to_num(ret)

        return ret

    @property
    def as_text(self) -> str:
        return self._code.decoded_text

    def __str__(self):
        return self.as_text

    def __repr__(self):
        LEN_LIMIT = 70
        code = str(self)
        code = code.replace("\n", "\\n")[:LEN_LIMIT] + (
            '' if len(code) <= LEN_LIMIT else '...'
        )
        return f"Code<\"{code}\">"

    @classmethod
    def get_from_file(
        cls,
        connection: 'Connection',
        path: 'Path | str',
        validate_code: bool = False,
    ) -> 'Code':
        """Create Code instance from contents of a local file.

        Note:
            Does not validate whether the path is correct, accessible or
            existing, so re-raises any OS-specific errors.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`.
            path (Path | str): Path to a file on your local drive containing
                the code.
                Example: `C:\\Program Files\\My Scripts\\some_script.py`
            validate_code (bool, optional): Whether to validate the code
                syntax upon creation. Defaults to False. (This is a pre-run
                validation only and does not guarantee no Runtime Errors.)

        Returns:
            Code: Created Code instance.
        """

        with open(Path(path)) as file:
            return cls(
                connection=connection,
                code=file.read(),
                validate_code=validate_code,
            )


class Script(
    Entity, DeleteMixin, MoveMixin, CopyMixin, CertifyMixin, ChangeJournalMixin
):
    """
    Class representing a Script object in Strategy.
    """

    _OBJECT_TYPE = ObjectTypes.SCRIPT
    _API_GETTERS = {
        **Entity._API_GETTERS,
        (
            "script_content",
            "script_result_type",
            "script_runtime_id",
            "script_type",
            # "script_usage_type",  # FYI: this is calculated from `subtype`
            "source",
            "variables",
        ): scripts_processor.get_info_data,
    }
    _API_GETTERS_KEEP_PRIVATE = {"variables", "script_content"}
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        "subtype": ObjectSubTypes,
        'certified_info': CertifiedInfo.from_dict,
        "owner": User.from_dict,
        "script_content": lambda source, connection: Code(connection, code=source),
        "script_result_type": ScriptResultType,
        "script_type": ScriptType,
        "script_usage_type": ScriptUsageType,
    }
    _API_PATCH = {
        **Entity._API_PATCH,
        (
            "name",
            "description",
            "folder_id",
            "script_type",
            "script_content",
            "script_runtime_id",
            "variables",
            "script_result_type",
        ): (scripts_processor.update_script, 'partial_put'),
    }

    _history_last_entry: dict = {}
    _variables_personal_answers: dict | None = None

    def __init__(
        self,
        connection: 'Connection',
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        """Initialize Script object by passing in either its ID or name.

        Args:
            connection (Connection): Strategy connection object returned by
                `connection.Connection()`.
            id (str, optional): ID of the Script.
            name (str, optional): Name of the Script.
        """

        connection._validate_project_selected()

        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'id' or 'name' parameter in the constructor."
                )

            opts = list_scripts(
                connection,
                project=connection.project_id,
                name=name,
                search_pattern=SearchPattern.EXACTLY,
            )

            if len(opts) != 1:
                raise ValueError(
                    f"Cannot uniquely identify Script with name '{name}'. "
                    f"Found {len(opts)} Scripts with this name. "
                    "Please provide ID instead."
                )

            id = opts[0].id

        super().__init__(connection, object_id=id)
        self.fetch()
        self._script_content._source_script = self

    def _init_variables(self, default_value=None, **kwargs):
        super()._init_variables(default_value, **kwargs)
        self._source: dict | None = kwargs.get('source', default_value)
        self._variables = (
            tuple(kwargs.get('variables')) if kwargs.get('variables') else default_value
        )
        self._script_content: Code = (
            Code(self._connection, kwargs.get('script_content'))
            if kwargs.get('script_content')
            else default_value
        )
        self._script_content._source_script = self
        self._script_type: ScriptType = (
            ScriptType(kwargs.get('script_type'))
            if kwargs.get('script_type')
            else default_value
        )
        self._script_runtime_id: str | None = kwargs.get(
            'script_runtime_id', default_value
        )
        self._script_result_type: ScriptResultType = (
            ScriptResultType(kwargs.get('script_result_type'))
            if kwargs.get('script_result_type')
            else default_value
        )

    @overload
    def execute(
        self,
        block_until_done: Literal[True] = True,
        pipe_logs: bool = False,
        raise_on_execution_failure: bool = False,
        variables_answers: VariablesAnswers | None = None,
        save_answers_as_personal: bool = False,
        allow_interactive_answering: bool = False,
    ) -> ExecutionStatus: ...

    @overload
    def execute(
        self,
        block_until_done: Literal[False] = False,
        variables_answers: VariablesAnswers | None = None,
        save_answers_as_personal: bool = False,
        allow_interactive_answering: bool = False,
    ) -> None: ...

    def execute(
        self,
        block_until_done: bool = True,
        pipe_logs: bool = False,
        raise_on_execution_failure: bool = False,
        variables_answers: VariablesAnswers | None = None,
        save_answers_as_personal: bool = False,
        allow_interactive_answering: bool = False,
    ) -> ExecutionStatus | None:
        """Start Script execution.

        Args:
            block_until_done (bool, optional): Whether to block the execution
                until it is done. Defaults to True.
            pipe_logs (bool, optional): Whether to pipe the execution logs
                to the logger while waiting for the execution to finish. Valid
                only if `block_until_done` is set. Defaults to False.
            raise_on_execution_failure (bool, optional): Whether to raise an
                exception if the execution fails. Valid only if
                `block_until_done` is set. Defaults to False.
            allow_interactive_answering (bool, optional): Whether to allow
                interactive answering of Variables if not all are answered.
                Defaults to False.

        Returns:
            ExecutionStatus: if `block_until_done` is set to True.
            Nothing otherwise.
        """

        variables = list(self.get_variables())

        if save_answers_as_personal:
            self.save_personal_variable_answers(variables_answers)

        return self._script_content.execute(
            variables=variables,
            answers=variables_answers,
            block_until_done=block_until_done,
            pipe_logs=pipe_logs,
            raise_on_execution_failure=raise_on_execution_failure,
            allow_interactive_answering=allow_interactive_answering,
        )

    def stop_execution(self) -> None:
        """Stops the Script execution."""

        return self._script_content.stop_execution()

    def get_execution_status(self) -> ExecutionStatus | None:
        """Retrieve current status of the Script execution.

        Returns:
            ExecutionStatus: Current status of the Script execution, or None if
                the Script is not currently running.
        """

        return self._script_content.get_current_execution_status()

    def wait_for_execution_finish(
        self, pipe_logs: bool = False, interval: int | None = None
    ) -> ExecutionStatus:
        """Wait until the Script execution is finished.

        Args:
            pipe_logs (bool, optional): Whether to pipe the execution logs of
                the Script to the console of this requester script. Defaults to
                False.
            interval (int, optional): Time interval in seconds between polling
                requests for status. If not provided, the value is taken from
                mstrio-py's `config`.

        Returns:
            ExecutionStatus: Final status of the Script execution.
        """

        return self._script_content.wait_for_execution_finish(
            pipe_logs=pipe_logs, interval=interval
        )

    def get_execution_details(self) -> dict:
        """Retrieve details about the Script current execution.

        Note:
            This property reads from Current Execution, hence refer to last run
            of the Script done by you specifically.

        Returns:
            dict: Details about the Script current execution.
        """

        return self._script_content.get_execution_details()

    def get_last_run_details(self) -> dict:
        """Gets the details about the last run of the Script, if any, including
        execution time, status and logs of the run.

        Note:
            This property reads from Run History, hence refer to last run of the
            Script overall, not necessarily your last run of this Script.

        Returns:
            dict: Dictionary with data about the last run of the Script,
                or empty dictionary if there is no history.
        """

        if self._script_content._evaluation_id:
            logger.info(
                "Script seems to be currently running so the data gathered may "
                "refer to a previous run, not current one."
            )

        self._history_last_entry = (
            scripts_processor.get_history(self.connection, self.id) or {}
        )

        return self._history_last_entry

    def _get_variable_factory(self) -> type['Variable']:
        match get_enum_val(
            self.script_usage_type, ScriptUsageType, throw_on_unknown=False
        ):
            case ScriptUsageType.DATASOURCE.value:
                return VariableDatasourceScript
            case ScriptUsageType.TRANSACTION.value:
                return VariableTransactionScript
            # TODO: add metric type when its Feature is ready
            case ScriptUsageType.STANDARD.value:
                return VariableStandardScript
            case _:
                raise ScriptSetupError(
                    "Unknown or unsupported Script Usage Type "
                    f"'{self.script_usage_type}'."
                )

    def _get_variables_personal_answers(self) -> dict[str, 'VariableAnswer']:
        if self._variables_personal_answers is None:
            try:
                self._variables_personal_answers = (
                    scripts_processor.get_variables_personal_answers(
                        self.connection, self.id
                    )
                )
            except (KeyError, ValueError):
                logger.warning(
                    "Getting Personal Default Answers for Variables failed. "
                    "Assuming none were set."
                )
                self._variables_personal_answers = {}

        return self._variables_personal_answers

    def get_variables(self) -> tuple['Variable']:
        """Get Variables saved in metadata of the Script.

        Returns:
            tuple[Variable]: Tuple of Variables associated with the Script.
        """

        Factory = self._get_variable_factory()

        # FYI: This is fine to recreate every time as usually there will be
        # around maybe 5 Variables. Even in edge cases there should be no more
        # than around 15. But even if there's 100, this is performant enough.
        variables = (
            tuple(
                (
                    # This keeps Variables types in sync with Script Usage Type.
                    # Below logic basically represents the following:
                    # 1. if an instance of exactly `Factory` class, keep it
                    # 2. build `Factory` class from...
                    #    a) if raw dict, from it directly
                    #    b) if an instance of other `Variable`, from its data
                    v
                    if isinstance(v, Factory)
                    else Factory.from_dict(
                        source=(
                            v.to_dict(skip_private_keys=True)
                            if isinstance(v, Variable)
                            else v
                        ),
                        connection=self._connection,
                    )
                )
                for v in self._variables
            )
            if self._variables
            else ()
        )

        for v in variables:
            v._source_script = self
            v._answer = None

        return variables

    def save_personal_variable_answers(
        self,
        answers: VariablesAnswers,
    ) -> None:
        """Save provided answers as personal default answers for Variables
        on the Script.

        Note:
            If some prompted Variables are not answered, current global default
            value for them will be used as your personal default. This is due to
            the REST API call requiring all prompted Variables to be provided
            at once.

        Args:
            answers (VariablesAnswers): Answers to be saved as personal
                default answers for Variables on the Script.
        """

        if not answers:
            raise ScriptSetupError("No answers were provided to be saved.")

        variables, answers = (
            self._script_content._convert_variables_and_answers_to_lists(
                variables=self.get_variables(),
                answers=answers,
                variables_factory=self._get_variable_factory(),
            )
        )

        if not all(
            isinstance(a, VariableAnswer) and a.is_valid() and a.source_variable.id
            for a in answers
        ):
            raise ScriptSetupError(
                "Some Variables or Variable answers provided to be saved as personal "
                "defaults are not valid. All provided answers need to be a valid "
                "Variable value for existing Variables and cannot be a flag to keep "
                "some default value."
            )

        answers = [a.to_dict() for a in answers]

        # FYI: API requires for all prompted variables to be provided in
        # payload. Below logic simulates not requiring that: by providing
        # global defaults if a variable is not answered.
        for var in (v for v in variables if v.prompt and not v.is_answered()):
            var.answer(keep_global_default=True)
            answers.append(var.get_answer_as_dict())

        scripts_processor.save_variables_personal_answers(
            self.connection,
            script_id=self.id,
            answers=answers,
        )

        # clear previous cache
        self._variables_personal_answers = None

    def add_variables(
        self,
        *data: 'Variable | dict',
    ) -> None:
        """Add Variables to the Script.

        Note:
            This method makes sure that provided data for new Variables is
            correct and valid before applying changes. If valid, saves to both
            metadata and locally.

        Args:
            *data (Variable | dict): Collected arbitrary amount of entries. For
                each Variable, an entry is either a Variable-class-based
                class instance or a dict with data to create a new Variable.

        Example:
        ```
            >>> script.add_variables(
            ...     {
            ...         "name": "var1",
            ...         "type": VariableType.TEXT,
            ...     },
            ...     VariableStandardScript(
            ...         name="var2",
            ...         type=VariableType.NUMERICAL,
            ...         prompt=True,
            ...     ),
            ... )
        ```
        """

        Factory = self._get_variable_factory()
        all_vars = self.get_variables()
        current_names = [v.name for v in all_vars]

        for v in data:
            if isinstance(v, dict):
                v = Factory.from_dict(v, self._connection)

            if not isinstance(v, Factory):
                raise ScriptSetupError(
                    f"Variable '{str(v)}' is not in valid shape. Make sure for "
                    "each new variable you provide a dict of valid data or "
                    f"instance of class {Factory.__name__}."
                )

            if v.name in current_names:
                raise ScriptSetupError(
                    f"Variable with name '{v.name}' already exists on {self}. Use "
                    "`Script.alter_variables(...)` method to edit an existing Variable."
                )

            current_names.append(v.name)

        self.alter(variables=list(all_vars) + list(data))

    def alter_variables(
        self,
        changes: dict[str, 'Variable | dict'] = None,  # FYI: `None` represent `{}`
    ) -> None:
        """Alter existing Variables on the Script.

        Note:
            This method makes sure that provided data for Variables to be
            altered is correct and valid before applying changes. If valid,
            saves to both metadata and locally.

        Args:
            changes (dict[str, Variable | dict]): Dictionary where keys are
                names or IDs of existing Variables to be altered, and values
                are either Variable-class-based class instances with new data
                or dicts with new data for the Variables.

        Example:
        ```
            >>> script.alter_variables({
            ...     "var1": {
            ...         "name": "new_variable_name",
            ...         "type": VariableType.TEXT,
            ...     },
            ...     "var2": VariableStandardScript(
            ...         name="another_new_name",
            ...         type=VariableType.NUMERICAL,
            ...         prompt=True,
            ...     ),
            ... })
        ```
        """

        # FYI: we allow not providing any changes if there were some changes
        # done locally to the existing Variables. Then, call to this method
        # just applies them to metadata
        changes = changes or {}

        Factory = self._get_variable_factory()
        all_vars = list(self.get_variables())
        current_names = [v.name for v in all_vars]

        for var_id_or_name, new_data in changes.items():
            target_var = Factory.get_variable_by_identifier(all_vars, var_id_or_name)

            if not target_var:
                raise ScriptSetupError(
                    f"Variable with identifier '{var_id_or_name}' does not exist on "
                    f"{self}. Cannot alter non-existing Variable."
                )

            if isinstance(new_data, Variable) and not isinstance(new_data, Factory):
                raise ScriptSetupError(
                    f"New data for Variable '{var_id_or_name}' is not in valid shape. "
                    "Make sure you provide either a dict of valid data or an instance "
                    f"of class {Factory.__name__}"
                )

            new_name = Variable._get_attr_from_variable(new_data, 'name')
            if new_name and new_name != target_var.name and new_name in current_names:
                raise ScriptSetupError(
                    f"Variable with name '{new_name}' already exists on {self}. "
                    "Cannot change to a name that is already in use by another "
                    "Variable."
                )

            current_names.append(new_name)

            idx = next(
                i
                for i, v in enumerate(all_vars)
                if v.name == var_id_or_name or v.id == var_id_or_name
            )
            all_vars[idx] = Factory.from_dict(
                {
                    **target_var.to_dict(),
                    **(new_data if isinstance(new_data, dict) else new_data.to_dict()),
                },
                self._connection,
            )

        self.alter(variables=all_vars)

    def delete_variables(
        self,
        *data: 'Variable | str',
    ) -> None:
        """Delete Variables from the Script.

        Note:
            This method makes sure that provided data for Variables to be
            deleted is correct and valid before applying changes. If valid,
            saves to both metadata and locally.

        Args:
            *data (Variable | str): Collected arbitrary amount of entries. For
                each Variable to be deleted, an entry is either a
                Variable-class-based class instance or the name/ID of the
                Variable as a string.

        Example:
        ```
            >>> script.delete_variables(
            ...     "var1",
            ...     VariableStandardScript(
            ...         name="var2",
            ...     ),
            ... )
        ```
        """

        if not all(isinstance(entry, (str, Variable)) for entry in data):
            raise ScriptSetupError(
                "At least one of provided arguments is not a valid type. "
                "All provided arguments must be either string representing name or ID "
                "of a Variable or Variable-based-class instance."
            )

        vars_for_remove = [
            entry.name if isinstance(entry, Variable) else entry for entry in data
        ]
        all_vars = self.get_variables()

        if not all(
            bool(Variable.get_variable_by_identifier(all_vars, var))
            for var in vars_for_remove
        ):
            raise ScriptSetupError(
                "At least one of provided arguments targets non-existing Variable. "
                "Please validate that you try to delete only existing Variables."
            )

        self.alter(
            variables=[
                v
                for v in all_vars
                if v.name not in vars_for_remove and v.id not in vars_for_remove
            ]
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        code: Code | str | None = None,
        validate_code: bool = False,
        # TODO: improve with resolver during Module for Runtimes dev
        runtime_id: str | None = None,
        variables: 'list[Variable | dict] | None' = None,
        script_type: ScriptType | str | None = None,
        script_result_type: ScriptResultType | str | None = None,
    ) -> None:
        """Alter Script's properties.

        Note:
            Script Usage Type cannot be changed in an existing Script. One needs
            to create a new Script to "change" the Usage type.

        Args:
            name (str, optional): New name of the Script.
            description (str, optional): New description of the Script.
            code (Code | str, optional): New code for the Script, either as
                Code class instance or as string (either already
                base64-encoded or original).
            validate_code (bool, optional): Whether to validate the code
                syntax before alteration. Defaults to False. Is ignored if code
                is not modified. (This is a pre-run validation only and does
                not guarantee no Runtime Errors.)
            runtime_id (str, optional): ID of the Script runtime.
            variables (list[Variable | dict], optional): List of Variables to
                be set for the Script.
            script_type (ScriptType | str, optional): Type of the Script.
            script_result_type (ScriptResultType | str, optional): Result type
                of the Script.
        """

        if code is not None:
            code = Code(self.connection, code, validate_code=validate_code)

        params = {
            'name': name,
            'description': description,
            'script_content': code._code.encoded_text if code else None,
            'script_runtime_id': runtime_id,
            'variables': (
                [v.to_dict() if isinstance(v, Variable) else v for v in variables]
                if variables is not None
                else None
            ),
            'script_type': get_enum_val(script_type, ScriptType),
            'script_result_type': get_enum_val(script_result_type, ScriptResultType),
        }
        expects_empty_vars = params.get("variables") == []
        params = delete_none_values(params, recursion=False)

        if expects_empty_vars:
            params["variables"] = []
            self._variables = None

        self._alter_properties(**params)

        if variables is not None:
            # remove cache if we are updating variables
            self._variables_personal_answers = None

    def _alter_properties(self, **properties):
        if new_fid := properties.get('folder_id'):
            self._folder_id = new_fid

        # FYI: REST requires some params in alter ALWAYS, even if not changed
        properties['name'] = properties.get('name', self.name)
        properties['folder_id'] = properties.get('folder_id', self.folder_id)

        return super()._alter_properties(**properties)

    def list_properties(self, excluded_properties=None):
        ret = super().list_properties(excluded_properties)

        if not excluded_properties or 'variables' not in excluded_properties:
            ret['variables'] = self.get_variables() or ()
            #                                       <>
            #                                      \__/

        return ret

    def to_dict(self, camel_case=True):  # NOSONAR # (missing params)
        ret = super().to_dict(camel_case=camel_case, skip_private_keys=True)

        content_key = 'scriptContent' if camel_case else 'script_content'
        ret[content_key] = str(ret.get(content_key, ''))

        if "variables" in ret:
            Factory = self._get_variable_factory()
            ret["variables"] = [
                # FYI: REST have some "shallow"-ly calculated defaults
                # in metadata. This parsing removes invalid edge cases
                Factory.from_dict(v, self.connection).to_dict()
                for v in ret["variables"]
            ]

        return ret

    @property
    def source(self) -> dict:
        return self._source

    @property
    def code(self) -> str:
        return str(self._script_content)

    @property
    def script_type(self) -> ScriptType:
        return self._script_type

    @property
    def script_runtime_id(self) -> str | None:
        return self._script_runtime_id

    # TODO: deliver this prop during Runtimes Module dev
    # @property
    # def script_runtime_details(self):
    #     pass

    @property
    def script_usage_type(self) -> ScriptUsageType:
        match self.subtype:
            case ObjectSubTypes.DATASOURCE_SCRIPT:
                return ScriptUsageType.DATASOURCE
            case ObjectSubTypes.TRANSACTION_SCRIPT:
                return ScriptUsageType.TRANSACTION
            case ObjectSubTypes.SCRIPT | ObjectSubTypes.COMMAND_MANAGER_SCRIPT:
                return ScriptUsageType.STANDARD
            # TODO: add metric types when their Feature is done
            case ObjectSubTypes.JUPYTER_NOTEBOOK_SCRIPT:
                raise ScriptSetupError(
                    "Script of type Jupyter Notebook is not supported."
                )
            case _:
                raise ScriptSetupError(
                    f'Unrecognized or unsupported Script Subtype: {self.subtype}'
                )

    @property
    def script_result_type(self) -> ScriptResultType:
        return self._script_result_type or ScriptResultType.UNKNOWN

    @property
    def last_execution_time(self) -> datetime | None:
        """Execution time as `datetime` object. `None` only if the Script was
        never run on server-side.

        Note:
            This property reads from Run History, hence refer to last run of the
            Script overall, not necessarily your last run of this Script.
        """

        value = self._history_last_entry.get('executionTime')
        return datetime.fromisoformat(value) if value else None

    @property
    def last_run_status(self) -> str | None:
        """Status of the last run of the Script (whether successful or not).
        `None` only if the Script was never run on server-side.

        Note:
            This property reads from Run History, hence refer to last run of the
            Script overall, not necessarily your last run of this Script.
        """

        return self._history_last_entry.get('lastStatus')

    @property
    def last_run_logs(self) -> str | None:
        """Logs from the last run of the Script. `None` only if the Script was
        never run on server-side.

        Note:
            This property reads from Run History, hence all log levels (stdout
            and stderr) are merged into one string and refer to last run of the
            Script overall, not necessarily your last run of this Script.
        """

        return self._history_last_entry.get('log')

    @property
    def execution_status(self) -> ExecutionStatus | None:
        """Status from last made current execution status request."""

        return self._script_content.status

    @property
    def execution_stdout(self) -> str:
        """STDOUT from last made current execution status request."""

        return self._script_content.stdout

    @property
    def execution_stderr(self) -> str:
        """STDERR from last made current execution status request."""

        return self._script_content.stderr

    @property
    def execution_pod_executor_message(self) -> str | None:
        """Message from Python Executor (if any) in last made current execution
        status request.

        Usually populated with error details when Code in Script failed during
        runtime.
        """

        return self._script_content.executor_message

    @property
    def execution_result(self) -> CodeResult:
        """Returned Value from last made current execution status request."""

        return self._script_content.result

    @classmethod
    def create(
        cls,
        connection: 'Connection',
        name: str,
        # TODO: improve with resolver during Module for Runtimes dev
        runtime_id: str,
        code: Code | str,
        destination_folder: 'Folder | tuple[str] | list[str] | str | None' = None,
        destination_folder_path: tuple[str] | list[str] | str | None = None,
        description: str | None = None,
        variables: 'list[Variable | dict] | None' = None,
        script_type: ScriptType | str = ScriptType.PYTHON,
        script_usage_type: ScriptUsageType | str = ScriptUsageType.STANDARD,
        script_result_type: ScriptResultType | str | None = None,
        validate_code: bool = False,
    ) -> 'Script':
        """Create a new Script in Strategy.

        Args:
            connection (Connection): Strategy connection object returned by
                `connection.Connection()`.
            name (str): Name of the Script.
            runtime_id (str): ID of the Script runtime.
            code (Code | str): Code class instance or code as a string, either
                base64-encoded or not
            destination_folder (Folder | tuple | list | str, optional): Folder
                object or ID or name or path specifying the folder where to
                create object.
            destination_folder_path (str, optional): Path of the folder.
                The path has to be provided in the following format:
                    /MicroStrategy Tutorial/Public Objects/Python Scripts
            description (str, optional): Description of the Script.
            variables (list[Variable | dict], optional): List of Variables'
                definitions (as raw dicts or as Variable class instances) to be
                saved in metadata of the Script.
            script_type (ScriptType | str, optional): Type of the Script.
                Defaults to ScriptType.PYTHON.
            script_usage_type (ScriptUsageType | str, optional): Usage type of
                the Script. Defaults to ScriptUsageType.STANDARD.
            script_result_type (ScriptResultType | str, optional): Result type
                of the Script. Defaults to None.
            validate_code (bool, optional): Whether to validate the code
                syntax before Script creation. Defaults to False. (This is a
                pre-run validation only and does not guarantee no Runtime
                Errors.)

        Returns:
            Script: Created Script class instance object.
        """

        connection._validate_project_selected()
        dest_id = get_folder_id_from_params_set(
            connection,
            connection.project_id,
            folder=destination_folder,
            folder_path=destination_folder_path,
        )

        new_obj_id = scripts_processor.create_script(
            connection=connection,
            name=name,
            runtime_id=runtime_id,
            encoded_code=Code(
                connection, code, validate_code=validate_code
            )._code.encoded_text,
            script_type=get_enum_val(script_type, ScriptType),
            folder_id=dest_id,
            description=description,
            variables=(
                [v.to_dict() if isinstance(v, Variable) else v for v in variables]
                if variables
                else None
            ),
            script_usage_type=get_enum_val(script_usage_type, ScriptUsageType),
            script_result_type=get_enum_val(script_result_type, ScriptResultType),
        )

        return Script(connection=connection, id=new_obj_id)


# --- Variables ---
_SYSTEM_PROMPTS_FOLDER_ID = "19BD7AE596A740C19341A88803DB87C2"


class SystemPrompt(Entity):
    """Simple representation of a System Prompt object."""

    _OBJECT_TYPE = ObjectTypes.PROMPT

    def __init__(self, connection: 'Connection', id: str):
        super().__init__(connection, object_id=id)

    @staticmethod
    def get_all(conn: 'Connection') -> list['SystemPrompt']:
        """Get all System Prompts available in the System Prompts Folder."""

        with config.temp_verbose_disable():
            return [
                SystemPrompt.from_dict(x, conn)
                for x in Folder(conn, id=_SYSTEM_PROMPTS_FOLDER_ID).get_contents(
                    to_dictionary=True
                )
            ]

    @staticmethod
    def get_by_name(conn: 'Connection', name: str) -> 'SystemPrompt':
        """Get System Prompt by its name."""

        try:
            return next(sp for sp in SystemPrompt.get_all(conn) if sp.name == name)
        except StopIteration:
            raise ScriptEnvironmentError(
                f"System Prompt with name '{name}' does not exist or "
                "cannot be accessed."
            )


class Variable(Dictable):
    """Script or Code Variable representation.

    Base Variable shape from which script-type-specific Variables inherit.
    """

    _FROM_DICT_MAP = {
        "type": VariableType,
        "object_ref": SystemPrompt.from_dict,
    }

    # descriptors
    name: str
    id: str | None = None
    type: 'VariableType | int' = VariableType.TEXT
    desc: str | None = None
    # shape (based on script type: which props are set may differ)
    multiple: bool | None = None
    prompt: bool | None = None
    transaction_column: bool | None = None
    nullable: bool | None = None
    required: bool | None = None
    editable: bool | None = None
    # actual answering and value
    value: VariableValue | None = None
    object_ref: 'SystemPrompt | dict | None' = None
    secret_value_input: bool | None = None

    _answer: 'VariableAnswer | FlagKeepDefaultAnswer | None' = None
    _source_script: 'Script | None' = None

    def __post_init__(self) -> None:
        """Post-calculation done after dataclass init.

        Note:
            Make sure to `super()`-call this in override methods.
        """
        # post-calculations
        # (required due to "shallow" REST defaults not covering edge cases)
        if self.transaction_column:
            if self.editable is None:
                self.editable = True

            # Falsy but not a list - reset to `None`
            if not self.value and not isinstance(self.value, list):
                self.value = None
        else:
            self.required = None
            self.nullable = None
            self.editable = None

        # validation
        if self.multiple:
            assert (self.value is None and self.prompt) or isinstance(
                self.value, list
            ), "When Variable is set to Multiple, its Value needs to be a list."
        else:
            assert (self.value is None and self.prompt) or not isinstance(
                self.value, list
            ), "When Variable is not set to Multiple, its Value cannot to be a list."

        is_multiple = self.multiple is True
        is_non_multi_type = self.type in [
            VariableType.DATE,
            VariableType.DATETIME,
            VariableType.SECRET,
            VariableType.SYSTEM_PROMPT,
        ]
        assert not (is_multiple and is_non_multi_type), (
            "Variables of type Date, Datetime, Secret or System Prompt cannot "
            "be set to Multiple."
        )

        if self.value is not None:
            assert VariableAnswer.is_of_valid_type(self, self.value), (
                "Value of the Variable is invalid. It does not correspond to "
                "the Variable's type properly."
            )

        assert self.type != VariableType.SYSTEM_PROMPT or (
            self.object_ref and not self.prompt
        ), (
            "System Prompt variable requires declared System Prompt to refer to "
            "and cannot be prompted."
        )

    def is_answered(self) -> bool:
        """Check whether the Variable has a valid answer.

        Returns:
            bool: True if the Variable is properly answered, False otherwise.
        """

        try:
            self.validate_whether_answered()
        except ScriptSetupError:
            return False
        return True

    def validate_whether_answered(self) -> None:
        """Validate whether the Variable is properly answered.
        Raises `ScriptSetupError` otherwise, with reasons.

        Raises:
            ScriptSetupError: When the Variable is not properly answered.
        """

        if not self.prompt:
            if self._answer:
                logger.warning(
                    f"Variable '{self.name}' is answered even though is not prompted. "
                    "Ignoring the invalid answer..."
                )
                self._answer = None

            if not VariableAnswer.is_of_valid_type(self, self.value):
                raise ScriptSetupError(
                    f"Variable '{self.name}' is not prompted and has invalid "
                    "or empty default value. Either make it prompted or set a "
                    "valid default value."
                )

            return

        if VariableAnswer.is_keep_default_flag(self._answer):
            value = (
                self.value
                if self._answer is VariableAnswer.KEEP_GLOBAL_DEFAULT
                else self.personal_default_answer
            )

            if (
                self.type != VariableType.SECRET
                and not VariableAnswer.is_of_valid_type(self, value)
            ):
                raise ScriptSetupError(
                    f"Variable '{self.name}' do not have a valid default value to "
                    "keep. Either make it prompted or set a valid default value."
                )

            return

        if not self._answer:
            raise ScriptSetupError(
                f"Variable '{self.name}' is prompted but not answered."
            )

        if not self._answer.is_valid():
            raise ScriptSetupError(
                f"Variable '{self.name}' has incorrect answer. Check whether types "
                "of Variable and its Answer match and whether the value is valid."
            )

    def answer(
        self,
        value: 'VariableValue | FlagKeepDefaultAnswer | None' = None,
        keep_personal_default: bool = False,
        keep_global_default: bool = False,
    ) -> 'VariableAnswer | FlagKeepDefaultAnswer':
        """Answer the Variable with the provided value (or a
        `FlagKeepDefaultAnswer` flag).

        Args:
            value (VariableValue | FlagKeepDefaultAnswer): Value to answer the
                Variable with or a flag to keep default value from a Variable.
                Optional only if `keep_default` flag is provided.
            keep_personal_default (bool): Whether to keep the personal default
                value from Variable as its answer. Defaults to False. If set to
                True, `value` parameter cannot be provided.
            keep_global_default (bool): Whether to keep the global default
                value from Variable as its answer. Defaults to False. If set to
                True, `value` parameter cannot be provided.

        Returns:
            VariableAnswer: Created answer for the Variable or a
                `FlagKeepDefaultAnswer` flag.
        """

        if keep_personal_default and keep_global_default:
            raise ScriptSetupError(
                "Cannot keep both personal and global default answer at the same time."
            )

        if not xor(value is not None, keep_personal_default or keep_global_default):
            raise ScriptSetupError(
                "Variable can either be answered with `value` or marked as "
                "`keep_<...>_default` but not both nor neither."
            )

        if keep_personal_default or value is VariableAnswer.KEEP_PERSONAL_DEFAULT:
            self._answer = VariableAnswer.KEEP_PERSONAL_DEFAULT
        elif keep_global_default or value is VariableAnswer.KEEP_GLOBAL_DEFAULT:
            self._answer = VariableAnswer.KEEP_GLOBAL_DEFAULT
        else:
            self._answer = VariableAnswer(self, value)

        self.validate_whether_answered()  # this will validate answer's structure
        return self._answer

    def get_answer_as_dict(self) -> dict:
        """Generate `to_dict` representation for REST execution request for
        Script class, where Script exists and have Variables definition stored
        in its metadata and we just need to provide actual answers.

        Note:
            This method assumes validation of the answer was already done.

        Returns:
            dict: Dictionary representation of the Variable's Answer.
        """

        if self._answer and not VariableAnswer.is_keep_default_flag(self._answer):
            return self._answer.to_dict()

        if self.type == VariableType.SECRET:
            # For SECRET we cannot gather true values from REST, so we need to
            # send empty value with a payload representing "use what you have
            # already in metadata"
            return VariableAnswer(self, value='', secret_value_input=False).to_dict()

        value = (
            self.value
            if self._answer is not VariableAnswer.KEEP_PERSONAL_DEFAULT
            else self.personal_default_answer
        )
        return VariableAnswer(self, value).to_dict()

    def get_answer_as_dict_for_non_script_run(self) -> dict:
        """Generate `to_dict` representation for REST execution request for
        Code class, where Script does not exist and full variable representation
        is required, non-prompted.

        It basically merges Variable definition and its Answer into one dict.

        Note:
            This method assumes validation of the answer was already done.

        Returns:
            dict: Dictionary representation of the Variable for non-Script run.
        """

        ret = self.to_dict(skip_private_keys=True)
        ret.update(self.get_answer_as_dict())
        ret['prompt'] = False
        ret['value'] = str(ret['value'])
        return ret

    @property
    def personal_default_answer(self) -> 'VariableValue':
        script = self._source_script
        vid = self.id

        if not vid or not script:
            raise ScriptSetupError(
                "Cannot gather or save personal answer to a Variable that is not "
                "saved in Script metadata on I-Server."
            )

        return script._get_variables_personal_answers().get(
            vid,
            # fall-back to global default if no personal answer was saved
            self.value,
        )

    @classmethod
    def from_dict(cls, *args, **kwargs):
        ret = super().from_dict(*args, **kwargs)
        if isinstance(cls, Variable):
            cls.__post_init__()  # NOSONAR
        return ret

    # TODO: consider improving typehint to all iterables, here and where
    # applicable after 3.10 is dropped (generics with params added in 3.11)
    @classmethod
    def get_variable_by_identifier(
        cls, variables: 'list[TVar]', id_or_name: str
    ) -> 'TVar | None':
        """Find Variable in the provided list by its ID or name.

        Args:
            variables (list[Variable | dict]): List of Variables to
                search in.
            id_or_name (str): ID or name of the Variable to find.

        Returns:
            Variable | dict | None: Found Variable or None if not found.
        """

        return next(
            (
                v
                for v in variables
                if (
                    Variable._get_attr_from_variable(v, 'id') == id_or_name
                    or Variable._get_attr_from_variable(v, 'name') == id_or_name
                )
            ),
            None,
        )

    @staticmethod
    def _get_attr_from_variable(variable: 'TVar', attr: str):
        if isinstance(variable, Variable):
            return getattr(variable, attr, None)

        if isinstance(variable, dict):
            return variable.get(attr)

        raise TypeError("Variable is neither Variable-class-based nor dict.")


TVar = TypeVar("TVar", Variable, dict)


@dataclass
class VariableStandardScript(Variable):
    """Simple representation of a Variable for Standard Scripts.

    Attributes:
        name: Variable name
        id: Variable ID
        type: Variable type
        desc: Variable description
        multiple: Whether the variable is a list of values
        prompt: Whether the variable prompts the user for input to be answered
        value: Variable default value, if any
        object_ref: For System Prompt variable, the System Prompt object
            referenced

    """

    name: str
    id: str | None = None
    type: 'VariableType | int' = VariableType.TEXT
    desc: str | None = None
    multiple: bool | None = False
    prompt: bool | None = True
    value: VariableValue | None = None
    object_ref: 'SystemPrompt | dict | None' = None

    def __post_init__(self) -> None:
        super().__post_init__()

        assert self.type != VariableType.TXN_ROW_PROVENANCE, (
            "For Standard Script, variable cannot be of type "
            "Transaction Provenance Column"
        )
        assert (
            isinstance(self.value, (str, int, float, date, datetime))
            # warning does not mention that secret will not be received from API
            # but it's by design
            or (self.id and self.type == VariableType.SECRET)
            or self.prompt
            or (self.object_ref and self.type == VariableType.SYSTEM_PROMPT)
        ), (
            "For Standard Script, variable needs to either have a value, "
            "be prompted or be a set System Prompt."
        )


@dataclass
class VariableDatasourceScript(Variable):
    """Simple representation of a Variable for Datasource Scripts.

    Attributes:
        name: Variable name
        id: Variable ID
        type: Variable type
        desc: Variable description
        multiple: Whether the variable is a list of values
        value: Variable default value, if any
    """

    name: str
    id: str | None = None
    type: 'VariableType | int' = VariableType.TEXT
    desc: str | None = None
    multiple: bool | None = False
    value: VariableValue | None = None

    def __post_init__(self):
        super().__post_init__()

        assert self.type not in [
            VariableType.TXN_ROW_PROVENANCE,
        ], (
            "For Datasource Script, variable cannot be of type "
            "Transaction Provenance Column"
        )


@dataclass
class VariableTransactionScript(Variable):
    """Simple representation of a Variable for Dashboard Transaction Scripts.

    Attributes:
        name: Variable name
        id: Variable ID
        type: Variable type
        desc: Variable description
        multiple: Whether the variable is a list of values
        prompt: Whether the variable prompts the user for input to be answered
        object_ref: For System Prompt variable, the System Prompt object
            referenced
        transaction_column: Whether the variable represents transaction column
        nullable: Whether the `transaction_column` variable can contain
            null values
        required: Whether the `transaction_column` variable is required to be
            answered in transaction dashboard
        editable: Whether the value of `transaction_column` variable in the
            transaction dashboard can be edited (or only added or removed)
    """

    name: str
    id: str | None = None
    type: 'VariableType | int' = VariableType.TEXT
    desc: str | None = None
    multiple: bool | None = False
    value: VariableValue | None = None
    prompt: bool | None = False
    object_ref: 'SystemPrompt | dict | None' = None
    transaction_column: bool | None = None
    nullable: bool | None = None
    required: bool | None = None
    editable: bool | None = None

    def __post_init__(self):
        super().__post_init__()

        if self.nullable or self.required or self.editable:
            assert self.transaction_column, (
                "Nullable, Required or Editable properties of a variable refer "
                "only to Transaction Column variable"
            )

        assert self.type != VariableType.TXN_ROW_PROVENANCE or (
            self.transaction_column
            and self.required
            and self.nullable is False
            and self.editable is False
        ), (
            "Transaction provenance Column variable needs to be explicitly set to "
            "Required and not Nullable and not Editable"
        )

        if self.transaction_column:
            assert self.value is None, "Transaction variable cannot have set value"
            assert (
                self.multiple and self.prompt
            ), "Transaction variable needs to be set to Multiple and Prompt"
            assert not (
                self.required and self.nullable
            ), "Required variable value cannot also be Nullable"
        else:
            assert (
                not self.prompt
            ), "For Transaction Script, non-transaction variables cannot be prompted"


# TODO: Finish and validate when the BHMO-105 Feature is released
# @dataclass
# class VariableMetricScript(Variable):
#     pass


@dataclass
class VariableAnswer(Variable):
    source_variable: 'Variable'
    value: VariableValue
    secret_value_input: bool = True

    def __post_init__(self):
        # FYI: explicitly not doing `super().__post_init__()`

        assert self.source_variable, "Cannot answer non-existent Variable"
        self.source_variable._answer = self

    def is_valid(self) -> bool:
        """Check whether the answer value is valid for the source variable.

        Returns:
            bool: True if the answer value is valid for the source variable.
        """

        is_non_prompted = self.source_variable.type in [
            VariableType.SYSTEM_PROMPT,
            VariableType.TXN_ROW_PROVENANCE,
        ]

        return not is_non_prompted and VariableAnswer.is_of_valid_type(
            self.source_variable, self.value
        )

    def to_dict(self, camel_case: bool = True):
        ret = super().to_dict(
            camel_case=camel_case,
            whitelist_keys=['value', 'secret_value_input'],
            skip_private_keys=True,
        )
        if vid := self.source_variable.id:
            ret['id'] = vid
        return ret

    @classmethod
    def is_of_valid_type(
        cls, source_variable: 'Variable', answer_value: VariableValue
    ) -> bool:
        """Check whether the provided answer value is of valid type for the
        source variable.

        Args:
            source_variable (Variable): Source Variable to check against.
            answer_value (VariableValue): Answer value to check.

        Returns:
            bool: True if the provided answer value is of valid type for the
                source variable.
        """

        if source_variable.multiple:
            if not isinstance(answer_value, (list, tuple)):
                return False

            mocked_var = deepcopy(source_variable)
            mocked_var.multiple = False

            return all(
                VariableAnswer.is_of_valid_type(mocked_var, value)
                for value in answer_value
            )

        is_number = source_variable.type == VariableType.NUMERICAL
        is_date = source_variable.type in [
            VariableType.DATE,
            VariableType.DATETIME,
        ]

        is_valid_number = is_number and isinstance(answer_value, (int, float))
        is_valid_date = is_date and is_valid_datetime(answer_value)
        is_valid_other = not is_date and not is_number and isinstance(answer_value, str)

        return is_valid_number or is_valid_date or is_valid_other

    @staticmethod
    def is_keep_default_flag(flag: Any) -> bool:
        return flag in [
            VariableAnswer.KEEP_GLOBAL_DEFAULT,
            VariableAnswer.KEEP_PERSONAL_DEFAULT,
        ]

    @staticmethod
    @property
    def KEEP_PERSONAL_DEFAULT() -> object:
        """Flag representing intention to keep personal-saved-default value of
        a Variable as an answer to the prompt-for-answer, instead of explicitly
        providing the answer.

        If no personal answer was saved, uses `KEEP_GLOBAL_DEFAULT` flag
        behavior instead.
        """
        # instance of singleton "nothing" to be compared via `is`
        return object()

    @staticmethod
    @property
    def KEEP_GLOBAL_DEFAULT() -> object:
        """Flag representing intention to keep global default value of a
        Variable as an answer to the prompt-for-answer, instead of explicitly
        providing the answer (even ignoring personal-saved-answer if it was
        set).

        Note:
            For `VariableType.SECRET` always defaults to `KEEP_PERSONAL_DEFAULT`
            if one is set.
        """
        # instance of singleton "nothing" to be compared via `is`
        return object()

    class For:
        """Helper method which exists solely to simplify answering prompted
        Variables in bulk.

        This allows to be used within `answers` params in `execute` methods in
        `Code` or `Script` to be provided as follows:

        ```
        >>> ...
        >>> answers=[
        >>>     VariableAnswer.For('var1').should_be('my-value'),
        >>>     VariableAnswer.For('var2').should_be_global_default,
        >>>     VariableAnswer.For('var3').should_be_personal_default,
        >>>     ...
        >>> ]
        >>> ...
        ```
        """

        def __init__(self, id_or_name: str):
            self.__id_or_name = id_or_name

        def should_be(self, value: 'VariableAnswer') -> 'AnswerForCallback':

            def __callback(
                variables: list['Variable'],
            ) -> 'VariableAnswer | FlagKeepDefaultAnswer':
                var = Variable.get_variable_by_identifier(variables, self.__id_or_name)

                if not var:
                    raise ScriptSetupError(
                        f"Variable with name or id '{self.__id_or_name}' does not "
                        "exist in available Variables collection."
                    )

                return var.answer(value)

            __callback._is_internal_mstrio = True

            return __callback

        @property
        def should_be_personal_default(self) -> 'AnswerForCallback':
            """Flag the Variable to keep personal default answer. If no such
            answer was saved, falls back to global default.
            """
            return self.should_be(VariableAnswer.KEEP_PERSONAL_DEFAULT)

        @property
        def should_be_global_default(self) -> 'AnswerForCallback':
            """Flag the Variable to keep global default answer.

            Note:
                For `VariableType.SECRET` always defaults to
                `should_be_personal_default` if one is set.
            """
            return self.should_be(VariableAnswer.KEEP_GLOBAL_DEFAULT)


# --- END: Variables ---



"""
This template shows how to create a Standard Script with return value.

When "Script Result Type" property is set, Standard Script can return a value.
- for Result Type `TEXT` you can return any string
- for Result Type `NUMERICAL` you can return a number (int or float)
- for Result Type `DATE` you can return a date or datetime string in ISO 8601 format

Below is a method required for the script to return a value shown multiple times
with different returning values. Use only one of those.
"""

# Script Result Type: TEXT
def get_results():
    return "This string will be returned from this script"

# OR

# Script Result Type: NUMERICAL
def get_results():
    return 42

# OR

# Script Result Type: DATE
def get_results():
    return "2024-01-15T10:30:00.000"



"""This is the demo script to show how to manage privileges and security roles for users and user groups.

In the first part privilege retrieval is presented, and in the second security role operations.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

# Privileges
from mstrio.access_and_security.privilege import Privilege
from mstrio.users_and_groups import list_user_groups, list_users, User, UserGroup
from mstrio.access_and_security.security_role import list_security_roles, SecurityRole
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
PRIVILEGE_NAME = $privilege_name  # Insert name of edited privilege here
PRIVILEGE_ID = $privilege_id  # Insert ID of edited privilege here

# get Privilege object by name
priv = Privilege(conn, name=PRIVILEGE_NAME)
# get Privilege object by id
priv = Privilege(conn, id=PRIVILEGE_ID)

# List Privileges and return objects or display in DataFrame
Privilege.list_privileges(conn, to_dataframe=True, is_project_level_privilege='True')
priv = Privilege.list_privileges(conn, id=[PRIVILEGE_ID])
for p in priv:
    print(p.id)
priv[0].list_properties()

# Define variables which can be later used in a script
USER_NAME = $user_name
USER_GROUP_NAME = $user_group_name
SECURITY_ROLE_NAME = $security_role_name  # Insert name of newly created or accessed security role
SECURITY_ROLE_DESCRIPTION = $security_role_description  # Insert description of newly created or accessed security role
SECURITY_ROLE_ID = $security_role_id
NEW_SECURITY_ROLE_NAME = $new_security_role_name
NEW_SECURITY_ROLE_DESCRIPTION = $new_security_role_description
NEW_SECURITY_ROLE_OWNER = $new_security_role_owner

# SecurityRoles
# List SecurityRoles and store the Objects
all_roles = list_security_roles(conn)
list_security_roles(conn, to_dataframe=True) # If True, return SecurityRoles as pandas DataFrame.
# List SecurityRoles by User
user = User(connection=conn, name=USER_NAME)
print(user.security_roles)
# List SecurityRoles by UserGroup
user_group = UserGroup(connection=conn, name=USER_GROUP_NAME)
print(UserGroup.security_roles)
# Get SecurityRole by name
security_role = SecurityRole(connection=conn, name=SECURITY_ROLE_NAME)
# also you can get SecurityRole by ID
security_role = SecurityRole(connection=conn, id=SECURITY_ROLE_ID)
# Create new SecurityRole
new_role = SecurityRole.create(
    conn,
    name=SECURITY_ROLE_NAME,
    description=SECURITY_ROLE_DESCRIPTION,
    privileges=[PRIVILEGE_ID, PRIVILEGE_NAME]
)
# Initialize SecurityRole object by ID
role = SecurityRole(conn, id=SECURITY_ROLE_ID)
# Initialize SecurityRole by name
role_by_name = SecurityRole(conn, name=SECURITY_ROLE_NAME)

# Alter SecurityRole. Change security role's name, description, or owner
role.alter(
    conn,
    name=NEW_SECURITY_ROLE_NAME,
    description=NEW_SECURITY_ROLE_DESCRIPTION,
    owner=NEW_SECURITY_ROLE_OWNER
)

# List SecurityRole members
role.list_members(project_name=PROJECT_NAME)

# Delete SecurityRole
security_role.delete(force=True) # If True, then no additional prompt will be shown before deleting.

# Define variables which can be later used in a script
USERNAME = $username  # Insert name of user to be assigned or revoked security role
USER_GROUP_NAME = $user_group_name  # Insert name of user group to be assigned or revoked security role

# Grant/Revoke Security Role to users/usergroups
user = User(conn, name=USERNAME)
users = list_users(conn)
group = UserGroup(conn, name=USER_GROUP_NAME)
groups = list_user_groups(conn)

# Grant/Revoke for Users
role.grant_to(members=user, project=PROJECT_NAME)
role.revoke_from(members=user, project=PROJECT_NAME)
role.grant_to(members=users, project=PROJECT_NAME)
role.revoke_from(members=users, project=PROJECT_NAME)

# Grant/Revoke for UserGroups
role.grant_to(members=group, project=PROJECT_NAME)
role.revoke_from(members=group, project=PROJECT_NAME)

# List Privileges
role.list_privileges()
role.list_privileges(to_dataframe=True)

# Grant/ Revoke privileges to Security Role
role.grant_privilege(privilege=[PRIVILEGE_ID])
role.revoke_privilege([PRIVILEGE_ID])
privs = list(role.list_privileges().keys())
role.revoke_all_privileges()
role.grant_privilege(privilege=privs)



import logging
from dataclasses import dataclass
from datetime import datetime
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import change_journal
from mstrio.connection import Connection
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors.change_journal import get_change_journals_loop
from mstrio.utils.version_helper import method_version_handler

if TYPE_CHECKING:
    from mstrio.server import Project
    from mstrio.utils.entity import Entity


logger = logging.getLogger(__name__)


@dataclass
class ChangeJournalEntry(Dictable):
    """Represents a single entry in the change journal."""

    project_id: str | None = None
    machine: str | None = None
    timestamp: str | None = None
    session_id: str | None = None
    transaction: dict | None = None
    user: dict | None = None
    changed_links: list[dict] | None = None
    changed_objects: list[dict] | None = None
    timestamp_iso: str | None = None


class TransactionType(AutoName):
    """Enumeration of possible transaction types in the change journal."""

    RESERVED = auto()
    ENABLE_CHANGE_JOURNAL = auto()
    DISABLE_CHANGE_JOURNAL = auto()
    PURGE_CHANGE_JOURNAL = auto()
    SAVE_OBJECT = auto()
    SAVE_OBJECTS = auto()
    SAVE_LINKITEMS = auto()
    DELETE_OBJECT = auto()
    DELETE_OBJECTS = auto()
    DELETE_PROJECT = auto()
    COPY_OBJECT = auto()
    MERGE_USER = auto()
    COLLECT_GARBAGE = auto()
    WRITE_SYSTEM_PROPERTY = auto()
    WRITE_DATABASE_PASSWORD = auto()


class ChangeType(AutoName):
    """Enumeration of possible change types in the change journal."""

    RESERVED = auto()
    CREATE_OBJECT = auto()
    CHANGE_OBJECT = auto()
    DELETE_OBJECT = auto()
    TOUCH_FOLDER = auto()
    CREATE_LINKITEM = auto()
    CHANGE_LINKITEM = auto()
    DELETE_LINKITEM = auto()
    CHANGE_JOURNAL_ENABLE = auto()
    CHANGE_JOURNAL_DISABLE = auto()
    CHANGE_JOURNAL_PURGE = auto()


@method_version_handler('11.4.0900')
def list_change_journal_entries(
    connection: Connection,
    transaction_sources: str | list[str] | None = None,
    transaction_types: (
        str | TransactionType | list[str | TransactionType] | None
    ) = None,
    change_types: str | ChangeType | list[str | ChangeType] | None = None,
    session_ids: str | list[str] | None = None,
    machines: str | list[str] | None = None,
    users: str | list[str] | None = None,
    affected_projects: str | list[str] | None = None,
    affected_objects: 'str | Entity | list[str | Entity] | None' = None,
    begin_transaction_id: int | None = None,
    end_transaction_id: int | None = None,
    begin_time: str | datetime | None = None,
    end_time: str | datetime | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
) -> list[ChangeJournalEntry]:
    """List change journal entries based on provided filters.

    Note: When listing all existing change journal entries across all projects,
        this function can take a considerable amount of time to execute,
        depending on the volume of data.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        transaction_sources (str | list[str], optional): Transaction sources to
            filter by. Can be a single string or list of strings.
        transaction_types (str | TransactionType | list[str | TransactionType],
            optional): Transaction types to filter by. Can be a single value or
            list of strings or TransactionType enums.
        change_types (str | ChangeType | list[str | ChangeType], optional):
            Change types to filter by. Can be a single value or list of strings
            or ChangeType enums.
        session_ids (str | list[str], optional): Session IDs to filter by. Can
            be a single session ID or list of session IDs.
        machines (str | list[str], optional): Machine names to filter by. Can
            be a single machine name or list of machine names.
        users (str | list[str], optional): User IDs to filter by. Can be a
            single user ID or list of user IDs.
        affected_projects (str | list[str], optional): Project IDs/names/objects
            to filter by. Can be a single element or list of elements.
        affected_objects (str | Entity | list[str | Entity], optional): Object
            IDs or Entity objects to filter by. Can be a single value or list of
            values.
        begin_transaction_id (int, optional): Start of the transaction ID range
            to filter by.
        end_transaction_id (int, optional): End of the transaction ID range to
            filter by.
        begin_time (str | datetime, optional): Start of the timestamp range to
            filter by. Format: 'MM/DD/YYYY HH:MM:SS AM/PM' for string input or a
            datetime object.
        end_time (str | datetime, optional): End of the timestamp range to
            filter by. Format: 'MM/DD/YYYY HH:MM:SS AM/PM' for string input or a
            datetime object.
        to_dictionary (bool, optional): If True, returns the results as a
            dictionary. Defaults to False.
        limit (int, optional): Maximum number of entries to return. If not
            specified, returns all matching entries.

    Returns:
        List of ChangeJournalEntry objects matching the filters.
    """

    from mstrio.utils.entity import Entity

    filter_criteria = {}

    if transaction_sources:
        filter_criteria['transactionSources'] = (
            transaction_sources
            if isinstance(transaction_sources, list)
            else [transaction_sources]
        )

    if transaction_types:
        if isinstance(transaction_types, (str, TransactionType)):
            transaction_types = [transaction_types]
        filter_criteria['transactionTypes'] = [
            item.value if isinstance(item, TransactionType) else item
            for item in transaction_types
        ]

    if change_types:
        if isinstance(change_types, (str, ChangeType)):
            change_types = [change_types]
        filter_criteria['changeTypes'] = [
            item.value if isinstance(item, ChangeType) else item
            for item in change_types
        ]

    if session_ids:
        filter_criteria['sessionIds'] = (
            session_ids if isinstance(session_ids, list) else [session_ids]
        )
    if machines:
        filter_criteria['machines'] = (
            machines if isinstance(machines, list) else [machines]
        )

    if users:
        filter_criteria['users'] = users if isinstance(users, list) else [users]

    if affected_projects:
        if not isinstance(affected_projects, list):
            affected_projects = [affected_projects]
        selected_projects = [
            get_project_id_from_params_set(
                connection,
                project=elem,
                no_fallback_from_connection=True,
            )
            for elem in affected_projects
        ]
        filter_criteria['affectedProjects'] = selected_projects

    if affected_objects:
        if not isinstance(affected_objects, list):
            affected_objects = [affected_objects]
        filter_criteria['affectedObjects'] = [
            {'id': obj.id} if isinstance(obj, Entity) else {'id': obj}
            for obj in affected_objects
        ]

    if begin_transaction_id:
        filter_criteria['beginTransactionId'] = begin_transaction_id

    if end_transaction_id:
        filter_criteria['endTransactionId'] = end_transaction_id

    if begin_time:
        if isinstance(begin_time, datetime):
            begin_time = begin_time.strftime("%m/%d/%Y %I:%M:%S %p")
        filter_criteria['beginTime'] = begin_time

    if end_time:
        if isinstance(end_time, datetime):
            end_time = end_time.strftime("%m/%d/%Y %I:%M:%S %p")
        filter_criteria['endTime'] = end_time

    with (
        config.temp_verbose_disable(),
        connection.temporary_project_change(project=None),
    ):
        search_id = (
            change_journal.create_change_journal_search_instance(
                connection, filter_criteria
            )
            .json()
            .get('searchId')
        )
        res = get_change_journals_loop(
            connection, search_id, offset=0, limit=limit or -1
        )

    if to_dictionary:
        return res
    return [ChangeJournalEntry(**entry) for entry in res]


@method_version_handler('11.4.0900')
def purge_change_journal_entries(
    connection: Connection,
    projects: 'list[str | Project] | Project | str | None' = None,
    comment: str | None = None,
    timestamp: str | datetime | None = None,
) -> None:
    """Purge change journal entries up to a specified timestamp for given
    projects.

    Note: If no projects are provided, all loaded projects will be targeted.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        projects (str | Project | list[str | Project]): Project(s) to purge
            change journal entries from. Can be a single project ID,
            project_name, Project object or a list of project
            IDs/names/objects.
        comment (str, optional): Comment for the purge operation.
        timestamp (str | datetime, optional): Timestamp up to which to purge
            entries. Format: 'MM/DD/YYYY HH:MM:SS AM/PM' for string input or a
            datetime object. If not provided, all entries will be purged.
    """

    if projects is None:
        projects = connection.environment.list_loaded_projects()

    if projects and not isinstance(projects, list):
        projects = [projects]
    selected_projects = [
        get_project_id_from_params_set(
            connection,
            project=elem,
            no_fallback_from_connection=True,
        )
        for elem in projects
    ]
    selected_projects = ",".join(selected_projects)

    if isinstance(timestamp, datetime):
        timestamp = _format_timestamp_for_api_purge(timestamp)

    res = change_journal.purge_change_journal_entries(
        connection, comment=comment, timestamp=timestamp, projects_ids=selected_projects
    )
    if config.verbose and res.ok:
        logger.info(
            "Request to purge change journal entries in selected projects was "
            "successfully sent."
        )


def _format_timestamp_for_api_purge(dt: datetime) -> str:
    """Format datetime for Strategy API timestamp requirements.

    Converts datetime to MM/DD/YYYY HH:MM:SS AM/PM format with proper
    constraints matching the API regex pattern. It requires to strip leading
    zeros from month, day, and hour components.

    Args:
        dt: Datetime object to format.

    Returns:
        Formatted timestamp string matching API requirements.
        Examples: '10/27/2025 4:37:20 PM', '1/5/2023 9:05:07 AM'
    """
    formatted = dt.strftime("%m/%d/%Y %I:%M:%S %p")

    parts = formatted.split(' ')
    date_part = parts[0]
    time_part = parts[1]
    ampm_part = parts[2]

    month, day, year = date_part.split('/')
    month = str(int(month))
    day = str(int(day))

    hour, minute, second = time_part.split(':')
    hour = str(int(hour))

    return f"{month}/{day}/{year} {hour}:{minute}:{second} {ampm_part}"



from typing import TYPE_CHECKING, Any

from mstrio.api import scripts as scripts_api
from mstrio.helpers import try_str_to_num
from mstrio.utils.encoder import Encoder
from mstrio.utils.helper import delete_none_values

if TYPE_CHECKING:
    from mstrio.connection import Connection


def _parse_variables_data_before_return(variables_data: list[dict]) -> None:
    # REST always returns value as string, even for `VariableType.NUMERICAL`
    for var in variables_data:
        if var.get('type') == 2 and isinstance(  # VariableType.NUMERICAL
            v := var.get('value'), str
        ):
            var['value'] = try_str_to_num(v)


def get_info_data(connection: 'Connection', id: str) -> dict:
    """Returns dict with data about a specific Script.

    Requires `connection` to have selected project.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): ID of the Script.

    Returns:
        dict: Dictionary with Script information.
    """
    data = scripts_api.get_info(
        connection=connection,
        project_id=connection.project_id,
        id=id,
    ).json()

    _parse_variables_data_before_return(data.get('variables', []))

    return data


def create_script(
    connection: 'Connection',
    name: str,
    runtime_id: str,
    encoded_code: str,
    script_type: str,
    folder_id: str,
    description: str | None = None,
    variables: list[dict] | None = None,
    script_usage_type: str | None = None,
    script_result_type: str | None = None,
) -> str:
    """Creates a new Script.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        name (str): Name of the Script.
        runtime_id (str): ID of the Script runtime.
        encoded_code (str): Base64 encoded content of the Script.
        script_type (str): Type of the Script.
        folder_id (str): ID of the folder where the Script will be created.
        description (str, optional): Description of the Script.
        variables (list[dict], optional): List of variables associated with the
            Script. Defaults to None.
        script_usage_type (str, optional): Usage type of the Script.
            Defaults to None, which means `commandManager` on I-Server.
        script_result_type (str, optional): Result type of the Script.
            Defaults to None which means "no result" on I-Server.

    Returns:
        str: ID of the newly created Script.
    """
    assert Encoder.is_encoded(
        encoded_code
    ), "Provided `encoded_code` is not b64-encoded."

    body = {
        "name": name,
        "description": description,
        "scriptRuntimeId": runtime_id,
        "scriptContent": encoded_code,
        "variables": variables,
        "scriptType": script_type,
        "scriptUsageType": script_usage_type,
        "scriptResultType": script_result_type,
        "folderId": folder_id,
    }
    body = delete_none_values(body, recursion=True)

    return scripts_api.create_script(
        connection,
        project_id=connection.project_id,
        body=body,
    ).json()['id']


def update_script(
    connection: 'Connection',
    id: str,
    body: dict,
) -> dict:
    """Updates an existing Script.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): ID of the Script to be updated.
        body (dict): Dictionary with updated Script information.

    Returns:
        dict: Dictionary with updated Script information after update.
    """

    if not body:
        return {}

    # FYI: this request returns something strange... we need fresh data to
    # return, requested independently
    res = scripts_api.update_script(
        connection,
        project_id=connection.project_id,
        id=id,
        body=body,
    )

    if not res.ok:
        return res

    return get_info_data(connection, id)


def delete_script(
    connection: 'Connection',
    script_id: str,
) -> bool:
    """Deletes a Script.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        script_id (str): ID of the Script to be deleted.

    Returns:
        bool: True if the Script was successfully deleted, False otherwise.
    """
    return scripts_api.delete_script(
        connection,
        project_id=connection.project_id,
        id=script_id,
    ).ok


def start_run(
    connection: 'Connection',
    script_id: str | None = None,
    encoded_code: str | None = None,
    runtime_id: str | None = None,
    variables_data: list[dict] | None = None,
) -> str:
    """Runs a Script or Code.

    Requires a set of parameters for either running script or running code,
    but not both. Those are either:
    - `script_id` + optional `variables_data`, or
    - `encoded_code` + `runtime_id` + optional `variables_data`.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        script_id (str, optional): ID of the Script to be run.
        encoded_code (str, optional): Base64 encoded content of the Script to be
            run.
        runtime_id (str, optional): ID of the Script runtime.
        variables_data (list[dict], optional): List of variables data to be used
            when running the Script.

    Returns:
        str: evaluation ID of the Script/Code execution job.
    """

    want_script: bool = script_id is not None
    want_code: bool = encoded_code is not None

    assert want_script ^ want_code, (
        "Incorrect set of parameters. Either `script_id` or `encoded_code` "
        "must be provided but not neither nor both."
    )

    if want_script:
        return scripts_api.run_existing_script(
            connection,
            project_id=connection.project_id,
            id=script_id,
            variables_data=variables_data,
        ).json()["id"]

    return scripts_api.run_python_code(
        connection,
        code=encoded_code,
        script_runtime_id=runtime_id,
        variables_data=variables_data,
    ).json()["id"]


def get_run_result(
    connection: 'Connection',
    evaluation_id: str,
) -> dict:
    """Gets results of a Script/Code execution.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        evaluation_id (str): Evaluation ID of the Script/Code execution job.

    Returns:
        dict: Dictionary with results of the execution.
    """
    return scripts_api.get_run_result(
        connection,
        evaluation_id=evaluation_id,
    ).json()


def stop_run(
    connection: 'Connection',
    evaluation_id: str,
) -> bool:
    """Stops a running Script/Code execution.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        evaluation_id (str): Evaluation ID of the Script/Code execution job.

    Returns:
        bool: True if the stop request was successful, False otherwise.
    """
    return scripts_api.stop_script_or_code_execution(
        connection,
        evaluation_id=evaluation_id,
    ).ok


def get_history(
    connection: 'Connection',
    script_id: str,
) -> dict | None:
    """Gets the history of Script's last run, if any.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        script_id (str): ID of the Script.

    Returns:
        dict | None: Dictionary with data about the last run of the Script,
            or None if there is no history.
    """

    # FYI: returns only last execution details
    items = scripts_api.get_last_run_results_in_bulk(
        connection,
        project_id=connection.project_id,
        script_ids=script_id,
    ).json()['history']

    return items[0] if items else None


def get_variables_personal_answers(
    connection: 'Connection',
    script_id: str,
) -> dict[str, Any]:
    """Gets personal answers for Script variables for the current user.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        script_id (str): ID of the Script.

    Returns:
        dict[str, Any]: Dictionary mapping variable IDs to their personal
            answers for current user.
    """

    data = scripts_api.get_script_variables_personal_answers(
        connection,
        project_id=connection.project_id,
        id=script_id,
    ).json()

    _parse_variables_data_before_return(data.get('answers', []))

    return {entry['id']: entry['value'] for entry in data['answers']}


def save_variables_personal_answers(
    connection: 'Connection',
    script_id: str,
    answers: list[dict],
) -> None:
    """Saves personal answers for Script variables for the current user.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`.
        script_id (str): ID of the Script.
        answers (list[dict]): List of dictionaries representing personal answers
            to be saved for current user.
    """

    scripts_api.update_script_variables_personal_answers(
        connection,
        project_id=connection.project_id,
        id=script_id,
        answers=answers,
    )



"""
    This is a template for creation of Aggregation Python Function which can be
    called in a Metric expression.

    It will work out of the box as is, when all variables are filled in.

    Method `get_results` needs to exist and return the values served as the
    metric evaluation results.

    Metric expression example: ScriptFunctionAggregation<_InputNames="lst, n",
                               _ScriptName=”nth_minimal_value”>( Cost, 2){~+}
"""

"""
    `lst` is a vector input
    `n` is a vector input
"""

def nth_minimal_value(lst, n):
    # Sort the list in ascending order
    sorted_lst = sorted(lst)

    # Check if n is within the range of the list length
    if n <= 0 or n > len(sorted_lst):
        raise ValueError("n is out of the valid range")

    # Return the nth minimal value (1-based index)
    return sorted_lst[n-1]

# must-have method which returns the values served as the metric evaluation results.
def get_results():
    return nth_minimal_value($lst, $n)

print(get_results())



import contextlib
import datetime as dt
from base64 import b64decode, b64encode

from pandas import DataFrame, Series
from pandas.api.types import is_datetime64_any_dtype


class Encoder:
    """Class handling base64 encoding and decoding of strings.

    As there are multiple instances of REST API requiring data to be encoded or
    being sent encoded so needed to be decoded, this helper class allows to do
    it consistently throughout mstrio-py.
    """

    _input: str = None
    _decoded_str: str = None
    _encoded_str: str = None

    def __init__(
        self,
        decoded_text: str | None = None,
        encoded_text: str | None = None,
    ) -> None:
        """Initialize Encoder object.

        Note:
            Provide either `encoded_text` or `decoded_text` but not both.

        Args:
            decoded_text (str, optional): Text to be encoded.
            encoded_text (str, optional): Text to be decoded.
        """
        if not bool(encoded_text) ^ bool(decoded_text):
            raise ValueError(
                "Provide either `encoded_text` or `decoded_text` but not "
                "both nor neither."
            )

        self._input = decoded_text or encoded_text

        if encoded_text:
            self._encoded_str = str(encoded_text)
            with contextlib.suppress(Exception):
                self._decoded_str = self.decode(self._encoded_str)
        else:
            self._decoded_str = str(decoded_text)
            try:
                temp = self.decode(self._decoded_str)
            except Exception:  # we expect fail
                temp = None
            finally:
                if temp:
                    # if did not fail and got value (aka successful decoding
                    # of allegedly already decoded text) -> fail
                    raise ValueError(
                        f"Provided decoded text '{self._decoded_str}' is "
                        "actually an encoded text"
                    )
            with contextlib.suppress(Exception):
                self._encoded_str = self.encode(self._decoded_str)

        if not self._decoded_str or not self._encoded_str:
            raise Exception(
                f"Provided text '{str(self._input)}' cannot be "
                f"properly {'decoded' if self._encoded_str else 'encoded'}"
            )

    @property
    def decoded_text(self) -> str:
        """Get decoded string."""
        return self._decoded_str

    @property
    def encoded_text(self) -> str:
        """Get encoded string."""
        return self._encoded_str

    @classmethod
    def is_encoded(cls, value: str) -> bool:
        """Check whether the provided string is base64-encoded."""
        if not value:
            return False
        try:
            return cls.encode(cls.decode(value)) == value
        except Exception:
            return False

    @staticmethod
    def decode(text: str) -> str:
        """Base64-decode a text string."""
        return b64decode(text.encode('utf-8')).decode('utf-8')

    @staticmethod
    def encode(text: str) -> str:
        """Base64-encode a text string."""
        return b64encode(text.encode('utf-8')).decode('utf-8')


class PandasDataframeEncoder:
    """Internal method for converting a Pandas DataFrame to Strategy One
    compliant base64 encoded JSON.

    When creating a data set, Strategy One APIs require the tabular data to
    have been transformed first into JSON and then into a base 64 encoded string
    before it is transmitted to the Intelligence Server via the REST API to
    create the data set. This class uses Pandas to handle transforming
    the DataFrame into a JSON representation of the data. For single-table data
    sets, Strategy One APIs require the JSON data to be formatted using
    the 'records' orientation from Pandas. Conversely, multi-table data sets
    require the JSON data to have a 'values' orientation. Based on the data set
    type, the correct encoding strategy is applied and the data is then encoded.

    Attributes:
        data_frame: Pandas DataFrame to be encoded.
        b64_data: DataFrame converted into a Base-64 encoded JSON string.
        orientation: For single-table data sets, "single"; for multi-table
            data sets, "multi".
    """

    def __init__(self, data_frame: DataFrame, dataset_type: str):
        """Inits Encoder with given data_frame and type.

        Args:
            data_frame (DataFrame): Pandas DataFrame to be converted.
            dataset_type (str): Dataset type. One of `single` or `multi` to
                correspond with single-table or multi-table sources.
        """
        self.data_frame: DataFrame = data_frame
        self._b64_data: str | None = None
        self.orientation: str | None = None
        # Mapping used when converting DataFrame rows
        # into proper JSON orientation needed for data uploads.
        _table_type_orient_map: dict = {'single': 'records', 'multi': 'values'}

        # Sets the proper orientation
        if dataset_type not in _table_type_orient_map:
            allowed_types = ', '.join(_table_type_orient_map)
            raise ValueError(f"Table type should be one of {allowed_types}")

        self.orientation = _table_type_orient_map[dataset_type]

    @staticmethod
    def _wrangle_date(x: Series) -> Series:
        if is_datetime64_any_dtype(x):
            return x
        if isinstance(x.iloc[0], dt.date):
            return x.astype('str')
        return x

    def encode(self) -> None:
        """Encode data in base 64."""
        self.data_frame = self.data_frame.apply(self._wrangle_date)

        json_data = self.data_frame.to_json(orient=self.orientation, date_format='iso')
        self._b64_data = Encoder(json_data).encoded_text

    @property
    def b64_data(self) -> str:
        if not self._b64_data:
            self.encode()
        return self._b64_data



"""This is the demo script to show how to manage Custom Groups. Its basic goal
is to present what can be done with the CustomGroup class, its related methods
and to ease its usage.

This script will not work without replacing parameters with real values.
"""

from mstrio.connection import get_connection
from mstrio.modeling.expression.enums import Function
from mstrio.modeling.expression.expression import Expression
from mstrio.modeling.expression.expression_nodes import AttributeFormPredicate
from mstrio.modeling.expression.parameters import (
    ConstantParameter,
    Variant,
    VariantType,
)
from mstrio.modeling.custom_group import (
    CustomGroup,
    CustomGroupElement,
    ElementDisplayOption,
    list_custom_groups,
)
from mstrio.modeling.schema.helpers import SchemaObjectReference

# Define variables which can be later used in a script
FOLDER_ID = $folder_id  # ID of the folder where the Custom Group will be created
SOURCE_CUSTOM_GROUP_ID = $source_custom_group_id  # ID of the Custom Group to copy elements from
CUSTOM_GROUP_NAME = $custom_group_name  # Name of the new Custom Group
CUSTOM_GROUP_NAME_UPDATED = $custom_group_name_updated  # Updated name of the Custom Group
QUALIFICATION_EXPRESSION_STRING = $qualification_expr_string  # Qualification expression string for the new element
ELEMENT_NAME_TO_DELETE = $element_name_to_delete  # Name of the element to delete from the Custom Group

# Connect to the Strategy One environment
conn = get_connection(workstationData, project_name="MicroStrategy Tutorial")

# List Custom Groups
listed = list_custom_groups(connection=conn)

source_cg = CustomGroup(conn, id=SOURCE_CUSTOM_GROUP_ID)

# Create Custom Group
new_cg = CustomGroup.create(
    connection=conn,
    name=CUSTOM_GROUP_NAME,
    destination_folder=FOLDER_ID,
    elements=source_cg.elements,
    drill_map_name="Empty Drill Map",
)

# List properties of Custom Group
properties = new_cg.list_properties()
print(properties)

# Alter Custom Group
print("Updating the new Custom Group...")
new_cg.alter(
    name=CUSTOM_GROUP_NAME_UPDATED,
)

# List elements of Custom Group
for element in new_cg.elements:
    print(f" - {element.name} (ID: {element.id})")

# Add element to Custom Group
print("Adding a new element to the Custom Group...")
new_element = new_cg.create_element(
    name="Element 4",
    display=ElementDisplayOption.ELEMENT,
    qualification=QUALIFICATION_EXPRESSION_STRING,
)

# Delete element from Custom Group
element_to_delete = new_cg.get_element(name=ELEMENT_NAME_TO_DELETE)
print(f"Deleting element '{element_to_delete.name}' from the Custom Group...")
new_cg.delete_element(id=element_to_delete.id)

# Delete Custom Group
new_cg.delete(force=True)
